"""module tests for the SAM learning algorithm"""

from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser, TrajectoryParser
from pddl_plus_parser.models import ActionCall, Observation, ObservedComponent, Predicate, Domain, Problem

from pytest import fixture
from sam_learning.learners import SAMLearner
from tests.consts import (
    sync_snapshot,
    extract_preconditions_predicates,
    BARMAN_CLASSICAL_DOMAIN_PATH,
    BARMAN_CLASSICAL_PROBLEM_PATH,
    BARMAN_CLASSICAL_TRAJECTORY_PATH,
    BARMAN_ERROR_IN_PRODUCTION_TRAJECTORY_PATH,
    BARMAN_ERROR_IN_PRODUCTION_DOMAIN_PATH,
    BARMAN_ERROR_IN_PRODUCTION_PROBLEM_PATH,
)
from utilities import NegativePreconditionPolicy


@fixture()
def barman_domain_production_bug() -> Domain:
    return DomainParser(BARMAN_ERROR_IN_PRODUCTION_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def barman_domain() -> Domain:
    return DomainParser(BARMAN_CLASSICAL_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def barman_problem(barman_domain: Domain) -> Problem:
    return ProblemParser(problem_path=BARMAN_CLASSICAL_PROBLEM_PATH, domain=barman_domain).parse_problem()


@fixture()
def barman_problem_production_bug(barman_domain_production_bug: Domain) -> Problem:
    return ProblemParser(
        problem_path=BARMAN_ERROR_IN_PRODUCTION_PROBLEM_PATH, domain=barman_domain_production_bug
    ).parse_problem()


@fixture()
def barman_observation(barman_domain: Domain, barman_problem: Problem) -> Observation:
    return TrajectoryParser(barman_domain, barman_problem).parse_trajectory(BARMAN_CLASSICAL_TRAJECTORY_PATH)


@fixture()
def barman_observation_production_bug(
    barman_domain_production_bug: Domain, barman_problem_production_bug: Problem
) -> Observation:
    return TrajectoryParser(barman_domain_production_bug, barman_problem_production_bug).parse_trajectory(
        BARMAN_ERROR_IN_PRODUCTION_TRAJECTORY_PATH
    )


@fixture()
def barman_sam_learning(barman_domain: Domain) -> SAMLearner:
    return SAMLearner(barman_domain, negative_preconditions_policy=NegativePreconditionPolicy.hard)


@fixture()
def barman_sam_production_bug(barman_domain_production_bug: Domain) -> SAMLearner:
    return SAMLearner(barman_domain_production_bug)


@fixture()
def elevators_sam_learning(elevators_domain: Domain) -> SAMLearner:
    return SAMLearner(elevators_domain)


@fixture()
def elevators_sam_learning_soft_policy(elevators_domain: Domain) -> SAMLearner:
    return SAMLearner(elevators_domain, negative_preconditions_policy=NegativePreconditionPolicy.soft)


@fixture()
def elevators_sam_learning_hard_policy(elevators_domain: Domain) -> SAMLearner:
    return SAMLearner(elevators_domain, negative_preconditions_policy=NegativePreconditionPolicy.hard)


def test_add_new_action_preconditions_adds_both_negative_and_positive_preconditions_to_the_action(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    test_action_call = observation_component.grounded_action_call
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    elevators_sam_learning._add_new_action_preconditions(grounded_action=test_action_call)
    learned_action_data = elevators_sam_learning.partial_domain.actions[test_action_call.name]
    discrete_preconditions = extract_preconditions_predicates(learned_action_data.preconditions)
    positive_preconditions = [precond for precond in discrete_preconditions if precond.is_positive]
    negative_preconditions = [precond for precond in discrete_preconditions if not precond.is_positive]
    assert len(positive_preconditions) > 0
    assert len(negative_preconditions) > 0
    print(learned_action_data.to_pddl())


def test_add_new_action_preconditions_adds_correct_positive_preconditions_to_action(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    test_action_call = observation_component.grounded_action_call
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    elevators_sam_learning._add_new_action_preconditions(grounded_action=test_action_call)
    learned_action_data = elevators_sam_learning.partial_domain.actions[test_action_call.name]
    discrete_preconditions = extract_preconditions_predicates(learned_action_data.preconditions)
    positive_preconditions = [
        precond.untyped_representation for precond in discrete_preconditions if precond.is_positive
    ]
    expected_conditions = {"(lift-at ?lift ?f1)", "(above ?f2 ?f1)", "(reachable-floor ?lift ?f2)"}
    assert expected_conditions.issubset(positive_preconditions)


def test_add_new_action_preconditions_do_not_adds_intersecting_positive_and_negative_preconditions(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    test_action_call = observation_component.grounded_action_call
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    elevators_sam_learning._add_new_action_preconditions(grounded_action=test_action_call)
    learned_action_data = elevators_sam_learning.partial_domain.actions[test_action_call.name]
    discrete_preconditions = extract_preconditions_predicates(learned_action_data.preconditions)
    positive_preconditions = {
        precond.untyped_representation for precond in discrete_preconditions if precond.is_positive
    }
    negative_preconditions = {
        precond.untyped_representation for precond in discrete_preconditions if not precond.is_positive
    }
    assert not positive_preconditions.intersection(negative_preconditions)


def test_update_action_preconditions_reduces_the_number_of_positive_predicates_but_does_not_remove_actual_preconditions(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call
    elevators_sam_learning.add_new_action(
        grounded_action=test_action_call, previous_state=previous_state, next_state=next_state
    )
    second_component = elevators_observation.components[4]
    sync_snapshot(elevators_sam_learning, second_component, elevators_observation.grounded_objects)
    second_test_action_call = second_component.grounded_action_call
    elevators_sam_learning._update_action_preconditions(grounded_action=second_test_action_call)
    discrete_preconditions = extract_preconditions_predicates(
        elevators_sam_learning.partial_domain.actions["move-down-slow"].preconditions
    )

    positive_preconditions = {
        precond.untyped_representation for precond in discrete_preconditions if precond.is_positive
    }
    negative_preconditions = {
        precond.untyped_representation for precond in discrete_preconditions if not precond.is_positive
    }
    assert positive_preconditions.issuperset(["(lift-at ?lift ?f1)", "(above ?f2 ?f1)", "(reachable-floor ?lift ?f2)"])
    assert all([p.startswith("(not") for p in negative_preconditions])


def test_update_action_preconditions_does_not_add_preconditions_not_observed_previously_when_first_added_the_action(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call
    test_partial_previous_state = previous_state.copy()
    test_partial_previous_state.state_predicates["(reachable-floor ?lift ?floor)"] = set()
    first_component = ObservedComponent(
        previous_state=test_partial_previous_state, next_state=next_state, call=test_action_call
    )
    sync_snapshot(elevators_sam_learning, first_component, elevators_observation.grounded_objects)

    elevators_sam_learning.add_new_action(
        grounded_action=test_action_call, previous_state=test_partial_previous_state, next_state=next_state
    )
    print(str(elevators_sam_learning.partial_domain.actions["move-down-slow"].preconditions))

    second_component = ObservedComponent(previous_state=previous_state, next_state=next_state, call=test_action_call)
    sync_snapshot(elevators_sam_learning, second_component, elevators_observation.grounded_objects)
    elevators_sam_learning._update_action_preconditions(test_action_call)
    discrete_preconditions = extract_preconditions_predicates(
        elevators_sam_learning.partial_domain.actions["move-down-slow"].preconditions
    )
    assert "(reachable-floor ?lift ?f2)" not in [precond.untyped_representation for precond in discrete_preconditions]


def test_update_action_preconditions_reduces_the_number_of_negative_predicates_but_does_not_remove_actual_preconditions(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)

    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call
    elevators_sam_learning.add_new_action(
        grounded_action=test_action_call, previous_state=previous_state, next_state=next_state
    )
    second_component = elevators_observation.components[4]
    sync_snapshot(elevators_sam_learning, second_component, elevators_observation.grounded_objects)
    second_test_action_call = second_component.grounded_action_call
    elevators_sam_learning._update_action_preconditions(grounded_action=second_test_action_call)
    discrete_preconditions = extract_preconditions_predicates(
        elevators_sam_learning.partial_domain.actions["move-down-slow"].preconditions
    )
    negative_preconditions = {
        precond.untyped_representation for precond in discrete_preconditions if not precond.is_positive
    }
    assert "(not (lift-at ?lift ?f2))" in negative_preconditions
    assert "(not (above ?f1 ?f2))" in negative_preconditions


def test_add_new_action_with_single_trajectory_component_adds_correct_preconditions_to_learned_action(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call

    elevators_sam_learning.add_new_action(
        grounded_action=test_action_call, previous_state=previous_state, next_state=next_state
    )

    added_action_name = "move-down-slow"
    assert added_action_name in elevators_sam_learning.partial_domain.actions
    learned_action_data = elevators_sam_learning.partial_domain.actions[added_action_name]
    discrete_preconditions = extract_preconditions_predicates(learned_action_data.preconditions)
    positive_preconditions = {
        precond.untyped_representation for precond in discrete_preconditions if precond.is_positive
    }
    negative_preconditions = {
        precond.untyped_representation for precond in discrete_preconditions if not precond.is_positive
    }
    assert positive_preconditions.issuperset(["(lift-at ?lift ?f1)", "(above ?f2 ?f1)", "(reachable-floor ?lift ?f2)"])
    assert all([p.startswith("(not") for p in negative_preconditions])


def test_add_new_action_with_single_trajectory_component_adds_correct_add_and_del_effects_to_the_action(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call
    elevators_sam_learning.add_new_action(
        grounded_action=test_action_call, previous_state=previous_state, next_state=next_state
    )

    added_action_name = "move-down-slow"
    assert added_action_name in elevators_sam_learning.partial_domain.actions
    learned_action_data = elevators_sam_learning.partial_domain.actions[added_action_name]
    assert {p.untyped_representation for p in learned_action_data.discrete_effects} == {
        "(lift-at ?lift ?f2)",
        "(not (lift-at ?lift ?f1))",
    }


def test_add_new_action_maintains_the_correct_add_and_delete_effects(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    elevators_sam_learning.add_new_action(test_action_call, previous_state, next_state)
    elevators_sam_learning.add_new_action(test_action_call, previous_state, next_state)
    learned_action = elevators_sam_learning.partial_domain.actions[test_action_call.name]
    assert {p.untyped_representation for p in learned_action.discrete_effects} == {
        "(lift-at ?lift ?f2)",
        "(not (lift-at ?lift ?f1))",
    }


def test_add_action_does_not_add_the_same_predicates_twice_to_the_preconditions(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call

    elevators_sam_learning.add_new_action(
        grounded_action=test_action_call, previous_state=previous_state, next_state=next_state
    )
    learned_action = elevators_sam_learning.partial_domain.actions[test_action_call.name]
    discrete_preconditions = extract_preconditions_predicates(learned_action.preconditions)
    positive_preconditions = [
        precond.untyped_representation for precond in discrete_preconditions if precond.is_positive
    ]
    negative_preconditions = [
        precond.untyped_representation for precond in discrete_preconditions if not precond.is_positive
    ]
    assert len(positive_preconditions) == len(set(positive_preconditions))
    assert len(negative_preconditions) == len(set(negative_preconditions))


def test_update_action_does_not_remove_effects_from_the_action(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call

    elevators_sam_learning.add_new_action(
        grounded_action=test_action_call, previous_state=previous_state, next_state=next_state
    )
    second_component = elevators_observation.components[4]
    sync_snapshot(elevators_sam_learning, second_component, elevators_observation.grounded_objects)
    second_previous_state = second_component.previous_state
    second_next_state = second_component.next_state
    second_test_action_call = second_component.grounded_action_call
    elevators_sam_learning.update_action(
        grounded_action=second_test_action_call, previous_state=second_previous_state, next_state=second_next_state
    )
    learned_action = elevators_sam_learning.partial_domain.actions[second_test_action_call.name]
    assert {p.untyped_representation for p in learned_action.discrete_effects} == {
        "(lift-at ?lift ?f2)",
        "(not (lift-at ?lift ?f1))",
    }


def test_update_action_does_not_add_the_same_predicates_twice_to_the_preconditions(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call

    elevators_sam_learning.add_new_action(
        grounded_action=test_action_call, previous_state=previous_state, next_state=next_state
    )
    elevators_sam_learning.update_action(
        grounded_action=test_action_call, previous_state=previous_state, next_state=next_state
    )
    learned_action = elevators_sam_learning.partial_domain.actions[test_action_call.name]
    discrete_preconditions = extract_preconditions_predicates(learned_action.preconditions)
    positive_preconditions = [
        precond.untyped_representation for precond in discrete_preconditions if precond.is_positive
    ]
    negative_preconditions = [
        precond.untyped_representation for precond in discrete_preconditions if not precond.is_positive
    ]
    assert len(positive_preconditions) == len(set(positive_preconditions))
    assert len(negative_preconditions) == len(set(negative_preconditions))


def test_update_action_with_two_trajectory_component_updates_action_data_correctly(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    first_observation_component = elevators_observation.components[0]
    second_observation_component = elevators_observation.components[4]

    elevators_sam_learning.current_trajectory_objects = elevators_observation.grounded_objects
    first_action_call = ActionCall(name="move-down-slow", grounded_parameters=["slow2-0", "n17", "n16"])
    second_action_call = ActionCall(name="move-down-slow", grounded_parameters=["slow1-0", "n9", "n8"])
    sync_snapshot(elevators_sam_learning, first_observation_component, elevators_observation.grounded_objects)
    elevators_sam_learning.add_new_action(
        grounded_action=first_action_call,
        previous_state=first_observation_component.previous_state,
        next_state=first_observation_component.next_state,
    )

    sync_snapshot(elevators_sam_learning, second_observation_component, elevators_observation.grounded_objects)
    elevators_sam_learning.update_action(
        grounded_action=second_action_call,
        previous_state=second_observation_component.previous_state,
        next_state=second_observation_component.next_state,
    )
    added_action_name = "move-down-slow"

    assert added_action_name in elevators_sam_learning.partial_domain.actions
    learned_action_data = elevators_sam_learning.partial_domain.actions[added_action_name]
    discrete_preconditions = extract_preconditions_predicates(learned_action_data.preconditions)
    preconditions_str = {p.untyped_representation for p in discrete_preconditions if p.is_positive}
    assert preconditions_str.issuperset(["(lift-at ?lift ?f1)", "(above ?f2 ?f1)", "(reachable-floor ?lift ?f2)"])
    assert {p.untyped_representation for p in learned_action_data.discrete_effects} == {
        "(lift-at ?lift ?f2)",
        "(not (lift-at ?lift ?f1))",
    }


def test_deduce_initial_inequality_preconditions_deduce_that_all_objects_with_same_type_should_not_be_equal(
    elevators_sam_learning: SAMLearner,
):
    elevators_sam_learning.should_enforce_injective_binding = True
    elevators_sam_learning.deduce_initial_inequality_preconditions()
    example_action_name = "move-up-slow"
    action = elevators_sam_learning.partial_domain.actions[example_action_name]
    assert action.preconditions.root.inequality_preconditions == {("?f1", "?f2")}


def test_verify_parameter_duplication_removes_inequality_if_found_action_with_duplicated_items_in_observation(
    elevators_sam_learning: SAMLearner,
):
    elevators_sam_learning.should_enforce_injective_binding = True
    elevators_sam_learning.deduce_initial_inequality_preconditions()
    example_action_name = "move-up-slow"
    action = elevators_sam_learning.partial_domain.actions[example_action_name]
    assert action.preconditions.root.inequality_preconditions == {("?f1", "?f2")}
    duplicated_action_call = ActionCall(name=example_action_name, grounded_parameters=["slow-lift", "c1", "c1"])
    elevators_sam_learning._verify_parameter_duplication(duplicated_action_call)
    assert len(action.preconditions.root.inequality_preconditions) == 0


def test_handle_action_effects_returns_delete_effects_with_predicates_with_is_positive_false_in_the_delete_effects_and_is_positive_true_in_the_add_effects(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    test_action_call = observation_component.grounded_action_call
    observed_objects = elevators_observation.grounded_objects
    elevators_sam_learning.current_trajectory_objects = observed_objects
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    add_effects, delete_effects = elevators_sam_learning._handle_action_effects(test_action_call)
    assert len(add_effects) > 0
    assert len(delete_effects) > 0
    for add_effects_predicate in add_effects:
        assert add_effects_predicate.is_positive

    for delete_effects_predicate in delete_effects:
        assert not delete_effects_predicate.is_positive


def test_handle_action_effects_returns_correct_add_and_delete_effects_predicates_after_action_is_executed(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    test_action_call = observation_component.grounded_action_call
    observed_objects = elevators_observation.grounded_objects
    elevators_sam_learning.current_trajectory_objects = observed_objects
    sync_snapshot(elevators_sam_learning, observation_component, elevators_observation.grounded_objects)
    add_effects, delete_effects = elevators_sam_learning._handle_action_effects(test_action_call)
    assert len(add_effects) == len(delete_effects) == 1
    assert add_effects[0].untyped_representation == "(lift-at ?lift ?f2)"
    assert delete_effects[0].untyped_representation == "(not (lift-at ?lift ?f1))"


def test_handle_action_effects_does_not_create_intersecting_sets_of_effects(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    previous_state = observation_component.previous_state
    next_state = observation_component.next_state
    test_action_call = observation_component.grounded_action_call
    observed_objects = elevators_observation.grounded_objects
    elevators_sam_learning.current_trajectory_objects = observed_objects
    add_effects, delete_effects = elevators_sam_learning._handle_action_effects(test_action_call)
    add_effects_str = set([p.untyped_representation for p in add_effects])
    delete_effects_str = set([p.untyped_representation for p in delete_effects])
    assert not add_effects_str.intersection(delete_effects_str)


def test_handle_single_trajectory_component_not_allowing_actions_with_duplicated_parameters(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    elevators_sam_learning.current_trajectory_objects = elevators_observation.grounded_objects
    test_action_call = ActionCall(name="move-down-slow", grounded_parameters=["slow2-0", "n17", "n17"])
    component = ObservedComponent(
        observation_component.previous_state, test_action_call, observation_component.next_state
    )
    elevators_sam_learning.handle_single_trajectory_component(component)

    added_action_name = "move-down-slow"
    learned_action_data = elevators_sam_learning.partial_domain.actions[added_action_name]
    assert len(learned_action_data.preconditions.root.operands) == 0
    assert len(learned_action_data.discrete_effects) == 0


def test_handle_single_trajectory_component_learns_preconditions_and_effects_when_given_a_non_duplicate_component(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    elevators_sam_learning.current_trajectory_objects = elevators_observation.grounded_objects
    elevators_sam_learning.handle_single_trajectory_component(observation_component)

    added_action_name = "move-down-slow"
    learned_action_data = elevators_sam_learning.partial_domain.actions[added_action_name]

    discrete_preconditions = extract_preconditions_predicates(learned_action_data.preconditions)
    discrete_preconditions_str = {p.untyped_representation for p in discrete_preconditions}
    assert discrete_preconditions_str.issuperset(
        ["(lift-at ?lift ?f1)", "(above ?f2 ?f1)", "(reachable-floor ?lift ?f2)"]
    )
    assert {p.untyped_representation for p in learned_action_data.discrete_effects} == {
        "(lift-at ?lift ?f2)",
        "(not (lift-at ?lift ?f1))",
    }


def test_handle_single_trajectory_component_maintains_the_correct_preconditions_invariants_after_multiple_executions(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    elevators_sam_learning.current_trajectory_objects = elevators_observation.grounded_objects
    elevators_sam_learning.handle_single_trajectory_component(observation_component)

    observation_component = elevators_observation.components[4]
    elevators_sam_learning.current_trajectory_objects = elevators_observation.grounded_objects
    elevators_sam_learning.handle_single_trajectory_component(observation_component)

    learned_action_data = elevators_sam_learning.partial_domain.actions[observation_component.grounded_action_call.name]
    discrete_preconditions = extract_preconditions_predicates(learned_action_data.preconditions)
    positive_preconditions = {p.untyped_representation for p in discrete_preconditions if p.is_positive}
    negative_preconditions = {p.untyped_representation for p in discrete_preconditions if not p.is_positive}
    assert positive_preconditions.issuperset(["(lift-at ?lift ?f1)", "(above ?f2 ?f1)", "(reachable-floor ?lift ?f2)"])

    assert negative_preconditions.issuperset(["(not (lift-at ?lift ?f2))", "(not (above ?f1 ?f2))"])
    print(learned_action_data.to_pddl())


def test_handle_single_trajectory_component_maintains_the_correct_effect_invariants_after_multiple_executions(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    elevators_sam_learning.current_trajectory_objects = elevators_observation.grounded_objects
    elevators_sam_learning.handle_single_trajectory_component(observation_component)

    observation_component = elevators_observation.components[4]
    elevators_sam_learning.current_trajectory_objects = elevators_observation.grounded_objects
    elevators_sam_learning.handle_single_trajectory_component(observation_component)

    learned_action_data = elevators_sam_learning.partial_domain.actions[observation_component.grounded_action_call.name]
    assert {p.untyped_representation for p in learned_action_data.discrete_effects} == {
        "(lift-at ?lift ?f2)",
        "(not (lift-at ?lift ?f1))",
    }


def test_handle_single_trajectory_component_maintains_the_correct_effect_and_does_not_miss_effect_predicates_on_production_bug(
    barman_sam_production_bug: SAMLearner, barman_observation_production_bug: Observation
):
    observation_component = barman_observation_production_bug.components[0]
    barman_sam_production_bug.current_trajectory_objects = barman_observation_production_bug.grounded_objects
    barman_sam_production_bug.handle_single_trajectory_component(observation_component)

    learned_action_data = barman_sam_production_bug.partial_domain.actions["refill_shot"]
    assert "(contains ?s ?i)" in [
        eff.untyped_representation for eff in learned_action_data.discrete_effects if eff.is_positive
    ]


def test_learn_action_model_does_not_obstruct_the_correctness_of_the_actions(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    learned_model, _ = elevators_sam_learning.learn_action_model([elevators_observation])

    learned_action_data = learned_model.actions[observation_component.grounded_action_call.name]
    assert {p.untyped_representation for p in learned_action_data.discrete_effects} == {
        "(lift-at ?lift ?f2)",
        "(not (lift-at ?lift ?f1))",
    }
    print(learned_action_data.to_pddl())


def test_learn_action_model_returns_learned_model(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    learned_model, learning_report = elevators_sam_learning.learn_action_model([elevators_observation])
    print(learning_report)
    print(learned_model.to_pddl())


def test_learn_action_model_with_hard_policy_deletes_negative_preconditions(
    elevators_sam_learning_hard_policy: SAMLearner, elevators_observation: Observation
):
    learned_model_ignore, _ = elevators_sam_learning_hard_policy.learn_action_model([elevators_observation])

    for action in learned_model_ignore.actions.values():
        for pre in action.preconditions.root.operands:
            if isinstance(pre, Predicate):
                assert pre.is_positive


def test_learn_action_model_with_hard_negative_precondition_policy_keep_positive_preconditions(
    elevators_sam_learning: SAMLearner,
    elevators_observation: Observation,
    elevators_sam_learning_hard_policy: SAMLearner,
):
    learned_model, _ = elevators_sam_learning.learn_action_model([elevators_observation])
    learned_model_ignore, _ = elevators_sam_learning_hard_policy.learn_action_model([elevators_observation])

    for action, action_ignored in zip(learned_model.actions.values(), learned_model_ignore.actions.values()):
        preconds = {prec for prec in action.preconditions.root.operands if isinstance(prec, Predicate)}
        preconds_ignore = {prec for prec in action_ignored.preconditions.root.operands if isinstance(prec, Predicate)}

        difference_ignore_from_classic = preconds_ignore.difference(preconds)
        difference_classic_from_ignore = preconds.difference(preconds_ignore)

        assert len(difference_ignore_from_classic) == 0

        for pre in difference_classic_from_ignore:
            if isinstance(pre, Predicate):
                assert not pre.is_positive


def test_learn_action_model_with_hard_policy_delete_effect_has_positive_precondition(
    elevators_sam_learning_hard_policy: SAMLearner, elevators_observation: Observation
):
    learned_model, _ = elevators_sam_learning_hard_policy.learn_action_model([elevators_observation])

    for action in learned_model.actions.values():
        predicates = [
            pre.untyped_representation
            for pre in action.preconditions.root.operands
            if pre.is_positive and isinstance(pre, Predicate)
        ]

        del_effects = [eff for eff in action.discrete_effects if not eff.is_positive and isinstance(eff, Predicate)]

        for del_eff in del_effects:
            del_eff.is_positive = not del_eff.is_positive

        flipped_del_effects = {del_eff.untyped_representation for del_eff in del_effects}

        assert len(del_effects) == len(flipped_del_effects.intersection(predicates))


def test_learn_action_model_with_soft_policy_delete_preconditions_has_add_effect(
    elevators_sam_learning: SAMLearner,
    elevators_sam_learning_soft_policy: SAMLearner,
    elevators_observation: Observation,
):
    learned_model, _ = elevators_sam_learning.learn_action_model([elevators_observation])
    learned_model_soft, _ = elevators_sam_learning_soft_policy.learn_action_model([elevators_observation])

    for action, action_ignored in zip(learned_model.actions.values(), learned_model_soft.actions.values()):
        preconds = {prec for prec in action.preconditions.root.operands if isinstance(prec, Predicate)}
        preconds_ignore = {prec for prec in action_ignored.preconditions.root.operands if isinstance(prec, Predicate)}

        difference_ignore_from_classic = preconds_ignore.difference(preconds)
        difference_classic_from_ignore = preconds.difference(preconds_ignore)
        add_effects = [effect.untyped_representation for effect in action.discrete_effects if effect.is_positive]

        assert len(difference_ignore_from_classic) == 0

        for pre in difference_classic_from_ignore:
            if isinstance(pre, Predicate):
                pre_positive_copy = pre.copy()
                pre_positive_copy.is_positive = True
                assert (not pre.is_positive) and (pre_positive_copy.untyped_representation in add_effects)


def test_learn_action_model_when_applying_multiple_times_with_different_trajectories_updates_the_learned_action_model_and_allows_learning_when_some_actions_are_deleted_since_they_are_not_being_observed(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    num_triplets = len(elevators_observation)
    observation1 = Observation()
    observation2 = Observation()
    observation1.add_problem_objects(elevators_observation.grounded_objects)
    observation2.add_problem_objects(elevators_observation.grounded_objects)
    for i in range(2):
        observation1.add_component(
            previous_state=elevators_observation.components[i].previous_state,
            call=elevators_observation.components[i].grounded_action_call,
            next_state=elevators_observation.components[i].next_state,
        )

    for i in range(2, num_triplets):
        observation2.add_component(
            previous_state=elevators_observation.components[i].previous_state,
            call=elevators_observation.components[i].grounded_action_call,
            next_state=elevators_observation.components[i].next_state,
        )
    learned_model1, learning_metadata = elevators_sam_learning.learn_action_model([observation1])
    num_learned_actions_model1 = len(learned_model1.actions)
    learned_model2, learning_metadata = elevators_sam_learning.learn_action_model([observation2])
    num_learned_actions_model2 = len(learned_model2.actions)
    assert num_learned_actions_model1 <= num_learned_actions_model2


def test_learn_action_model_when_applying_multiple_times_with_different_trajectories_updates_the_learned_action_model_and_allows_learning_and_creates_a_more_concise_action_model(
    elevators_sam_learning: SAMLearner, elevators_observation: Observation
):
    num_triplets = len(elevators_observation)
    half_components = num_triplets // 2
    observation1 = Observation()
    observation2 = Observation()
    observation1.add_problem_objects(elevators_observation.grounded_objects)
    observation2.add_problem_objects(elevators_observation.grounded_objects)
    for i in range(half_components):
        observation1.add_component(
            previous_state=elevators_observation.components[i].previous_state,
            call=elevators_observation.components[i].grounded_action_call,
            next_state=elevators_observation.components[i].next_state,
        )

    for i in range(half_components, num_triplets):
        observation2.add_component(
            previous_state=elevators_observation.components[i].previous_state,
            call=elevators_observation.components[i].grounded_action_call,
            next_state=elevators_observation.components[i].next_state,
        )
    learned_model1, learning_metadata = elevators_sam_learning.learn_action_model([observation1])
    model1_str = learned_model1.to_pddl()
    num_learned_actions_model1 = len(learned_model1.actions)
    learned_model2, learning_metadata = elevators_sam_learning.learn_action_model([observation2])
    model2_str = learned_model2.to_pddl()
    num_learned_actions_model2 = len(learned_model2.actions)
    assert len(model1_str) > len(model2_str) or num_learned_actions_model1 < num_learned_actions_model2


def test_learn_action_model_when_learning_barman_domain_returns_correct_domain_and_no_missing_predicates_are_found_in_the_preconditions(
    barman_sam_learning: SAMLearner, barman_observation: Observation
):
    learned_model1, learning_metadata = barman_sam_learning.learn_action_model([barman_observation])
    assert "(holding ?h1 ?s)" in learned_model1.actions["shake"].to_pddl()



"""Module tests for the numeric function matcher."""
from pddl_plus_parser.models import Domain, PDDLFunction, ActionCall, Observation
from pytest import fixture

from sam_learning.core import NumericFunctionMatcher
from tests.consts import (
    TRUCK_TYPE,
    FUEL_COST_FUNCTION,
    CURRENT_LOAD_GROUNDED_TRAJECTORY_FUNCTION,
    LOAD_LIMIT_GROUNDED_TRAJECTORY_FUNCTION,
)

TEST_NUMERIC_LOAD_LIMIT_FUNCTION = PDDLFunction(name="load_limit", signature={"?t": TRUCK_TYPE})

TEST_GROUNDED_NUMERIC_LOAD_LIMIT_FUNCTION = PDDLFunction(name="load_limit", signature={"tru1": TRUCK_TYPE})

CURRENT_LIMIT_TRAJECTORY_FUNCTION = PDDLFunction(name="current_load", signature={"truck1": TRUCK_TYPE})


@fixture()
def numeric_function_matcher(depot_domain: Domain) -> NumericFunctionMatcher:
    return NumericFunctionMatcher(domain=depot_domain)


@fixture()
def minecraft_function_matcher(minecraft_domain: Domain) -> NumericFunctionMatcher:
    return NumericFunctionMatcher(domain=minecraft_domain)


def test_create_possible_function_signatures_with_simple_action_creates_all_possible_signature_permutations(
    numeric_function_matcher: NumericFunctionMatcher,
):
    test_action_call_params = ActionCall(name="load", grounded_parameters=["hoist1", "crate1", "tru1", "loc1"])
    possible_signatures = numeric_function_matcher.create_possible_function_signatures(
        action_parameters=test_action_call_params.parameters, numeric_function=TEST_NUMERIC_LOAD_LIMIT_FUNCTION
    )

    assert len(possible_signatures) == 4
    assert possible_signatures == [
        "(load_limit hoist1)",
        "(load_limit crate1)",
        "(load_limit tru1)",
        "(load_limit loc1)",
    ]


def test_create_possible_function_signatures_when_action_is_called_with_a_constant_returns_all_possible_combinations_of_grounded_functions(
    minecraft_function_matcher: NumericFunctionMatcher, minecraft_domain: Domain
):
    test_action_call_params = ActionCall(name="tp_to", grounded_parameters=["crafting_table", "cell767"])
    tp_to_functions = PDDLFunction(name="cell_type", signature={"crafting_table": minecraft_domain.types["cell"]})
    possible_signatures = minecraft_function_matcher.create_possible_function_signatures(
        action_parameters=test_action_call_params.parameters, numeric_function=tp_to_functions
    )
    assert len(possible_signatures) == 2


def test_lift_matched_parameters_when_given_function_with_constant_that_can_match_multiple_options_returns_all_options(
    minecraft_function_matcher: NumericFunctionMatcher, minecraft_domain: Domain
):
    tp_to_function = PDDLFunction(name="cell_type", signature={"crafting_table": minecraft_domain.types["cell"]})
    grounded_execution = ["crafting_table"]
    lifted_functions = minecraft_function_matcher.lift_matched_parameters(
        executed_action=minecraft_domain.actions["tp_to"],
        grounded_call_parameters=grounded_execution,
        grounded_function=tp_to_function,
    )
    assert len(lifted_functions) == 2


def test_lift_matched_parameters_creates_lifted_version_of_function(
    numeric_function_matcher: NumericFunctionMatcher, depot_domain: Domain
):
    TEST_GROUNDED_NUMERIC_LOAD_LIMIT_FUNCTION.set_value(132.87)
    test_lifted_action = depot_domain.actions["load"]
    test_action_call_params = ActionCall(name="load", grounded_parameters=["hoist1", "crate1", "tru1", "loc1"])
    lifted_function = numeric_function_matcher.lift_matched_parameters(
        executed_action=test_lifted_action,
        grounded_call_parameters=test_action_call_params.parameters,
        grounded_function=TEST_GROUNDED_NUMERIC_LOAD_LIMIT_FUNCTION,
    )[0]

    assert lifted_function.untyped_representation == "(load_limit ?z)"
    assert lifted_function.value == 132.87


def test_match_state_functions_with_single_parameterized_grounded_function_finds_all_possible_matches(
    numeric_function_matcher: NumericFunctionMatcher,
):
    LOAD_LIMIT_GROUNDED_TRAJECTORY_FUNCTION.set_value(411.0)
    FUEL_COST_FUNCTION.set_value(34.0)
    simple_state_fluents = {
        "(fuel-cost )": FUEL_COST_FUNCTION,
        "(load_limit truck1)": LOAD_LIMIT_GROUNDED_TRAJECTORY_FUNCTION,
    }
    test_action_call = ActionCall(name="load", grounded_parameters=["hoist2", "crate1", "truck1", "distributor1"])
    matches = numeric_function_matcher.match_state_functions(
        action_call=test_action_call, grounded_state_fluents=simple_state_fluents
    )
    assert len(matches) == 2
    assert list(matches.keys()) == ["(load_limit ?z)", "(fuel-cost )"]
    assert matches["(fuel-cost )"].value == 34.0
    assert matches["(load_limit ?z)"].value == 411.0


def test_match_state_functions_with_small_number_of_grounded_functions_finds_all_possible_matches(
    numeric_function_matcher: NumericFunctionMatcher,
):
    LOAD_LIMIT_GROUNDED_TRAJECTORY_FUNCTION.set_value(411.0)
    CURRENT_LOAD_GROUNDED_TRAJECTORY_FUNCTION.set_value(121.0)
    FUEL_COST_FUNCTION.set_value(34.0)

    simple_state_fluents = {
        "(fuel-cost )": FUEL_COST_FUNCTION,
        "(load_limit truck1)": LOAD_LIMIT_GROUNDED_TRAJECTORY_FUNCTION,
        "(current_load truck1)": CURRENT_LOAD_GROUNDED_TRAJECTORY_FUNCTION,
    }
    test_action_call = ActionCall(name="load", grounded_parameters=["hoist2", "crate1", "truck1", "distributor1"])
    matches = numeric_function_matcher.match_state_functions(
        action_call=test_action_call, grounded_state_fluents=simple_state_fluents
    )
    assert len(matches) == 3
    assert list(matches.keys()) == ["(load_limit ?z)", "(current_load ?z)", "(fuel-cost )"]
    assert matches["(fuel-cost )"].value == 34.0
    assert matches["(load_limit ?z)"].value == 411.0
    assert matches["(current_load ?z)"].value == 121.0


def test_match_state_functions_matches_from_actual_trajectory_state(
    numeric_function_matcher: NumericFunctionMatcher, depot_observation: Observation
):
    observation_component = depot_observation.components[0]
    test_action_call = observation_component.grounded_action_call
    test_previous_state_fluents = observation_component.previous_state.state_fluents
    possible_matches = numeric_function_matcher.match_state_functions(
        action_call=test_action_call, grounded_state_fluents=test_previous_state_fluents
    )

    for matched_lifted_function in possible_matches.values():
        for parameter in matched_lifted_function.signature:
            assert parameter in ["?x", "?y", "?z"]



from sam_learning.core.numeric_learning.convex_hull_learner import ConvexHullLearner
from .environment_snapshot import EnvironmentSnapshot
from .exceptions import NotSafeActionError
from .learner_domain import LearnerAction, LearnerDomain
from .learning_types import EquationSolutionType, ConditionType
from sam_learning.core.numeric_learning.linear_regression_learner import LinearRegressionLearner
from .matching_utils import (
    extract_discrete_effects,
    contains_duplicates,
    create_signature_permutations,
    extract_not_effects
)
from sam_learning.core.numeric_learning.numeric_fluent_learner_algorithm import NumericFluentStateStorage
from sam_learning.core.numeric_learning.numeric_function_matcher import NumericFunctionMatcher
from .numeric_learning import (
    construct_multiplication_strings,
    prettify_coefficients,
    prettify_floating_point_number,
    construct_linear_equation_string,
    construct_non_circular_assignment,
)
from .predicates_matcher import PredicatesMatcher
from .vocabulary_creator import VocabularyCreator



"""Creates the trajectories that will be used in the trajectory"""
import logging
import sys
from pathlib import Path

from pddl_plus_parser.exporters import MetricFFParser, TrajectoryExporter, ENHSPParser
from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser

from utilities import SolverType


class ExperimentTrajectoriesCreator:
    """Class responsible for creating the trajectories that will be used in the experiments."""

    domain_file_name: str
    working_directory_path: Path
    logger: logging.Logger

    def __init__(self, domain_file_name: str, working_directory_path: Path):
        self.domain_file_name = domain_file_name
        self.working_directory_path = working_directory_path
        self.logger = logging.getLogger(__name__)

    def fix_solution_files(self, solver_algorithm: SolverType = SolverType.enhsp) -> None:
        """Fix the solution files created by the solvers so that the learning algorithms will be able to use.

        :param solver_algorithm: the solver that was used to create the plans.
        """
        if solver_algorithm == SolverType.metric_ff:
            solution_parser = MetricFFParser()
            for solution_file_path in self.working_directory_path.glob("*.solution"):
                self.logger.debug(f"Fixing the solution file - {solution_file_path.stem}")
                solution_parser.parse_plan(solution_file_path, solution_file_path)

        elif solver_algorithm == SolverType.enhsp:
            solution_parser = ENHSPParser()
            for solution_file_path in self.working_directory_path.glob("*.solution"):
                self.logger.debug(f"Lowering the case for all actions in - {solution_file_path.stem}")
                solution_parser.parse_plan(solution_file_path)

    def create_domain_trajectories(self) -> None:
        """Creates the domain trajectory files."""
        domain_file_path = self.working_directory_path / self.domain_file_name
        domain = DomainParser(domain_file_path).parse_domain()
        trajectory_exporter = TrajectoryExporter(domain=domain)
        for solution_file_path in self.working_directory_path.glob("*.solution"):
            problem_file_path = self.working_directory_path / f"{solution_file_path.stem}.pddl"
            trajectory_file_path = self.working_directory_path / f"{solution_file_path.stem}.trajectory"
            if trajectory_file_path.exists():
                continue

            problem = ProblemParser(problem_path=problem_file_path, domain=domain).parse_problem()
            try:
                triplets = trajectory_exporter.parse_plan(problem, solution_file_path)
                trajectory_exporter.export_to_file(triplets, trajectory_file_path)

            except (ValueError, IndexError):
                continue


if __name__ == "__main__":
    trajectory_creator = ExperimentTrajectoriesCreator(domain_file_name=sys.argv[1], working_directory_path=Path(sys.argv[2]))
    selected_solver = SolverType.enhsp
    trajectory_creator.fix_solution_files(selected_solver)
    trajectory_creator.create_domain_trajectories()



"""Tests for numeric_utils.py"""

import numpy as np
import pandas as pd
from sklearn.decomposition import PCA

from sam_learning.core.numeric_learning.numeric_utils import (
    extract_numeric_linear_coefficient,
    construct_projected_variable_strings,
    extended_gram_schmidt,
    create_monomials,
    create_polynomial_string,
    reduce_complementary_conditions_from_convex_hull,
    remove_complex_linear_dependencies,
)


def test_extract_numeric_linear_coefficient_does_not_output_infinity_for_zero_division():
    """Test that the function does not output infinity for zero division."""
    try:
        extract_numeric_linear_coefficient(pd.Series([1, 2, 3]), pd.Series([0, 0, 0]))
    except ZeroDivisionError:
        assert False


def test_extract_numeric_linear_coefficient_outputs_correct_coefficient_when_division_is_legal():
    """Test that the function outputs the correct coefficient when the division is legal."""
    assert extract_numeric_linear_coefficient(pd.Series([1, 2, 3]), pd.Series([1, 2, 3])) == 1


def test_construct_projected_variable_strings_returns_correct_strings_representing_the_projected_format():
    function_variables = ["(x)", "(y)", "(z)"]
    X = np.array([[-1, -1, 1], [-2, -1, 3], [-3, -2, 23], [1, 1, 31], [2, 1, 12], [3, 2, 7]])
    pca = PCA(n_components=2)
    pca.fit(X)

    pca_variable_strings = construct_projected_variable_strings(function_variables, pca.mean_, pca.components_)
    print(pca_variable_strings)
    assert len(pca_variable_strings) == 2  # 2 components


def test_construct_projected_variable_strings_returns_correct_strings_representing_the_projected_format_example_for_nsam_paper():
    function_variables = ["(x)", "(y)", "(z)", "w"]
    X = np.array([[1, 0, 17, 3], [5, 5, 7, 3], [2, 4, 9, 7]])
    shift = np.array([-1, 2, 3, 12])

    pca_variable_strings = construct_projected_variable_strings(function_variables, shift, X)
    print(f"x' = {pca_variable_strings[0]}")
    print(f"y' = {pca_variable_strings[1]}")
    print(f"z' = {pca_variable_strings[2]}")

    diagonal_eye = [list(vector) for vector in np.eye(X.shape[1])]
    orthnormal_span = extended_gram_schmidt(diagonal_eye, [[1, 0, 17, 3], [5, 5, 7, 3], [2, 4, 9, 7]])
    print(f"orthnormal_span = {orthnormal_span}")


def test_construct_projected_variable_strings_returns_correct_strings_representing_the_projected_format_when_components_are_equal_to_variables():
    function_variables = ["(x)", "(y)", "(z)"]
    X = np.array([[-1, -1, 1], [-2, -1, 3], [-3, -2, 23], [1, 1, 31], [2, 1, 12], [3, 2, 7]])
    pca = PCA(n_components=3)
    pca.fit(X)

    pca_variable_strings = construct_projected_variable_strings(function_variables, pca.mean_, pca.components_)
    assert len(pca_variable_strings) == 3  # 3 components
    print(pca_variable_strings)


def test_construct_projected_variable_strings_returns_correct_numbers_according_to_construction_calculation_for_simple_case():
    function_variables = ["(x)", "(y)", "(z)"]
    mean = np.array([2, 2, 2])
    pca_components = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    pca_variable_strings = construct_projected_variable_strings(function_variables, mean, pca_components)
    assert len(pca_variable_strings) == 3  # 3 components
    assert pca_variable_strings[0] == "(- (x) 2)"
    assert pca_variable_strings[1] == "(- (y) 2)"
    assert pca_variable_strings[2] == "(- (z) 2)"
    print(pca_variable_strings)


def test_construct_pca_variable_strings_returns_correct_numbers_according_to_construction_calculation_for_complex_case():
    function_variables = ["(x)", "(y)", "(z)"]
    mean = np.array([2, 2, 2])
    pca_components = np.array([[1, 8, 32.5], [12, 1, 33], [3, 2, 1]])

    pca_variable_strings = construct_projected_variable_strings(function_variables, mean, pca_components)
    assert len(pca_variable_strings) == 3  # 3 components
    assert pca_variable_strings[0] == "(+ (- (x) 2) (+ (* (- (y) 2) 8) (* (- (z) 2) 32.5)))"
    assert pca_variable_strings[1] == "(+ (* (- (x) 2) 12) (+ (- (y) 2) (* (- (z) 2) 33)))"
    assert pca_variable_strings[2] == "(+ (* (- (x) 2) 3) (+ (* (- (y) 2) 2) (- (z) 2)))"


def test_extended_gram_schmidt_on_no_shift_base_returns_correct_lower_dimensional_points():
    projections = extended_gram_schmidt([[0, 0, 0], [0, 1, 0], [1, 0, 0]])
    assert len(projections) == 2
    assert projections == [[0, 1, 0], [1, 0, 0]]


def test_extended_gram_schmidt_on_no_shift_base_returns_correct_lower_dimensional_points_with_correct_rational_values():
    projections = extended_gram_schmidt([[0, 0, 0], [0, 1, 1], [0, 1, -1]])
    normalized_projection = []
    for vector in projections:
        normalized_projection.append([round(val, 3) for val in vector])

    assert len(projections) == 2
    assert normalized_projection == [[0, 0.707, 0.707], [0, 0.707, -0.707]]


def test_extended_gram_schmidt_on_no_shift_when_given_orthonormal_partial_base_returns_complementary_base_to_points():
    projections = extended_gram_schmidt([[1, 0, 0], [0, 1, 0], [0, 0, 1]], [[0, 1, 0], [1, 0, 0]])
    assert len(projections) == 1
    assert projections == [[0, 0, 1]]


def test_extended_gram_schmidt_on_no_shift_with_base_correctly_returns_the_missing_vector():
    projections = extended_gram_schmidt([[0, 0, 0], [0, 1, 1], [0, 1, -1]])
    extended_standard_base = np.concatenate((np.eye(3), np.array(projections)), axis=0).tolist()
    complementary_projection = extended_gram_schmidt(extended_standard_base, projections)
    assert len(complementary_projection) == 1
    assert complementary_projection == [[1, 0, 0]]


def test_extended_gram_schmidt_on_no_shift_with_base_correctly_returns_the_missing_vector_with_more_complicated_setting():
    projections = extended_gram_schmidt([[0, 0, 0], [0, 1, 1], [0, 1, -1]])
    extended_standard_base = [[18, 2, 19], [2, 7, 1], [29, 31, -1]]
    complementary_projection = extended_gram_schmidt(extended_standard_base, projections)
    assert len(complementary_projection) == 1
    assert complementary_projection == [[1.0, -9.868649107779169e-17, 0.0]]


def test_extended_gram_schmidt_on_no_shift_with_base_with_more_rows_than_columns_still_returns_2x2_output():
    projections = extended_gram_schmidt([[-19.0, 32.0], [14.0, 52.0], [28.0, 12.0], [-7.0, 13.0]])
    assert len(projections) == 2
    assert all([len(vector) == 2 for vector in projections])
    assert np.sum(np.array(projections) - np.array([[-0.510538754155436, 0.859854743840735], [0.859854743840735, 0.510538754155436]])) < 0.0001
    print(projections)


def test_extended_gram_schmidt_when_one_variable_is_a_constant_extract_the_constant_and_return_orthonormal_base():
    # no need to add a non-zero constant since subtracting from the first sample will result in a zero vector
    projections = extended_gram_schmidt([[8, 2, 0], [72, 33, 0], [11, 8, 0], [1, 1, 0], [17, 13, 0]])
    assert len(projections) == 2


def test_extended_gram_schmidt_with_no_base_returns_correct_orthonormal_matrix():
    projections = extended_gram_schmidt([[1, 1, 1], [2, 1, 0], [5, 1, 3]])
    assert len(projections) == 3
    assert (
        np.sum(
            np.array(projections)
            - np.array(
                [
                    [0.577350269189626, 0.577350269189626, 0.577350269189626],
                    [0.707106781186548, 0, -0.707106781186548],
                    [0.408248290463863, -0.816496580927726, 0.408248290463863],
                ]
            )
        )
        < 0.0001
    )


def test_extended_gram_schmidt_with_paper_example_returns_correct_values():
    projections = extended_gram_schmidt([[0, 0, 0], [-1, 1, 0], [-1, 0, 1]])
    print(projections)
    assert len(projections) == 2


def test_reduce_complementary_conditions_from_convex_hull_basis_can_eliminate_values_when_computing_complementary_basis_with_constant_value():
    test_values = [[np.random.randint(0, 100), np.random.randint(0, 100), 0] for _ in range(10)]
    projections = extended_gram_schmidt(test_values)
    diagonal_eye = [list(vector) for vector in np.eye(3)]
    complementary_basis = extended_gram_schmidt(diagonal_eye, projections)
    result = reduce_complementary_conditions_from_convex_hull(projections, complementary_basis)
    assert result == projections


def test_reduce_complementary_conditions_from_convex_hull_basis_can_eliminate_values_when_computing_complementary_basis_with_linear_dependent_value():
    test_values = [[i, 2 * i, np.random.randint(0, 100)] for i in range(10)]
    projections = extended_gram_schmidt(test_values)
    diagonal_eye = [list(vector) for vector in np.eye(3)]
    complementary_basis = extended_gram_schmidt(diagonal_eye, projections)
    result = reduce_complementary_conditions_from_convex_hull(projections, complementary_basis)
    for index, projection in enumerate(result):
        assert projection[0] == 0
        assert projection[1] == 0.5 * projections[index][0] + projections[index][1]


def test_create_monomials_creates_correct_monomials_when_given_polynom_degree_one():
    test_functions = ["(fuel-cost )", "(load_limit ?z)", "(current_load ?z)"]
    monomials = create_monomials(test_functions, polynom_degree=1)
    assert len(monomials) == 3 + 3
    assert sum([len(monomial) for monomial in monomials]) == 3 + 3 * 2


def test_create_monomials_creates_correct_monomials_when_given_polynom_degree_zero():
    test_functions = ["(fuel-cost )", "(load_limit ?z)", "(current_load ?z)"]
    monomials = create_monomials(test_functions, polynom_degree=0)
    assert len(monomials) == 3


def test_create_monomials_creates_correct_monomials_when_given_polynom_degree_two():
    test_functions = ["(fuel-cost )", "(load_limit ?z)", "(current_load ?z)"]
    monomials = create_monomials(test_functions, polynom_degree=2)
    assert len(monomials) == 3 + 3 + 3
    print(monomials)


def test_create_polynomial_string_when_calling_with_a_single_function_returns_the_function_string():
    monomial = ["(fuel-cost )"]
    polynomial_string = create_polynomial_string(monomial)
    assert polynomial_string == "(fuel-cost )"


def test_create_polynomial_string_when_calling_with_two_functions_returns_the_multiplication_of_the_two_functions():
    monomial = ["(fuel-cost )", "(load_limit ?z)"]
    polynomial_string = create_polynomial_string(monomial)
    assert polynomial_string == "(* (fuel-cost ) (load_limit ?z))"


def test_create_polynomial_string_when_calling_with_three_functions_returns_the_multiplication_of_the_three_functions():
    monomial = ["(fuel-cost )", "(load_limit ?z)", "(current_load ?z)"]
    polynomial_string = create_polynomial_string(monomial)
    assert polynomial_string == "(* (fuel-cost ) (* (load_limit ?z) (current_load ?z)))"


def test_remove_complex_linear_dependencies_returns_correct_conditions_and_reduced_dataframe_when_one_feature_in_the_dataframe_is_constant_value():
    test_data = pd.DataFrame({"a": [1, 1, 1, 1], "b": [1, 3, 3, 4], "c": [1, 2, 3, 4]})
    reduced_data, conditions = remove_complex_linear_dependencies(test_data)
    assert len(conditions) == 1
    assert conditions[0] == "(= a 1)"
    assert len(reduced_data.columns) == 2
    assert "a" not in reduced_data.columns


def test_remove_complex_linear_dependencies_returns_correct_conditions_and_reduced_dataframe_when_one_feature_equals_the_assignment_of_two_others_and_a_const():
    test_data = pd.DataFrame({"a": [1, 2, 3, 4], "b": [1, 3, 3, 4], "c": [3, 6, 7, 9]})
    reduced_data, conditions = remove_complex_linear_dependencies(test_data)
    assert len(conditions) == 1
    assert conditions[0] == "(= a (+ (* b -1) (+ c -1)))"
    assert len(reduced_data.columns) == 2
    assert "a" not in reduced_data.columns


def test_remove_complex_linear_dependencies_returns_correct_conditions_and_reduced_dataframe_when_one_feature_equals_the_assignment_of_two_others_and_a_const_and_there_is_a_const_column():
    test_data = pd.DataFrame({"a": [1, 2, 3, 4], "b": [1, 3, 3, 4], "c": [3, 6, 7, 9], "d": [1, 1, 1, 1]})
    reduced_data, conditions = remove_complex_linear_dependencies(test_data)
    assert len(conditions) == 2
    assert set(conditions) == {"(= a (+ (* b -1) (+ c -1)))", "(= d 1)"}
    assert len(reduced_data.columns.tolist()) == 2
    assert "a" not in reduced_data.columns.tolist()


def test_remove_complex_linear_dependencies_when_all_columns_are_constant_values_returns_empty_dataframe_and_the_pddl_constraint_that_checks_the_column_name_is_equal_to_the_constant():
    test_data = pd.DataFrame({"a": [1, 1, 1, 1], "b": [1, 1, 1, 1], "c": [1, 1, 1, 1]})
    reduced_data, conditions = remove_complex_linear_dependencies(test_data)
    assert len(conditions) == 3
    assert conditions == ["(= a 1)", "(= b 1)", "(= c 1)"]
    assert len(reduced_data.columns) == 0



"""Utility functionality for the experiments"""
from pathlib import Path
from typing import List, Union
from typing import Optional

from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.lisp_parsers import ProblemParser, TrajectoryParser

from statistics.ma_performance_calculator import MASamPerformanceCalculator
from statistics.numeric_performance_calculator import NumericPerformanceCalculator
from statistics.semantic_performance_calculator import SemanticPerformanceCalculator
from utilities import LearningAlgorithmType

DEFAULT_SIZE = 10


NUMERIC_ALGORITHMS = [
    LearningAlgorithmType.numeric_sam,
    LearningAlgorithmType.naive_nsam,
    LearningAlgorithmType.naive_polysam,
    LearningAlgorithmType.incremental_nsam,
    LearningAlgorithmType.plan_miner,
]

MULTI_AGENT_ALGORITHMS = [
    LearningAlgorithmType.ma_sam,
    LearningAlgorithmType.ma_sam_plus,
    LearningAlgorithmType.sam_learning,
]


def init_semantic_performance_calculator(
    working_directory_path: Path,
    domain_file_name: Optional[str] = None,
    domain_file_path: Optional[Path] = None,
    learning_algorithm: LearningAlgorithmType = LearningAlgorithmType.sam_learning,
    executing_agents: Optional[List[str]] = None,
    test_set_dir_path: Path = None,
    problem_prefix: str = "pfile",
) -> Union[NumericPerformanceCalculator, SemanticPerformanceCalculator, MASamPerformanceCalculator]:
    """Initializes a numeric performance calculator object.

    :param working_directory_path: the directory path where the domain and problem files are located.
    :param domain_file_name: the name of the domain file.
    :param domain_file_path: the path to the domain file.
    :param learning_algorithm: the type of learning algorithm to use.
    :param executing_agents: the agents that are executing the domain.
    :param test_set_dir_path: the path to the directory containing the test set.
    :param problem_prefix: the prefix of the problem files.
    :return: the initialized numeric performance calculator object.
    """
    if domain_file_name is None and domain_file_path is None:
        raise ValueError("Either 'domain_file_name' or 'domain_file_path' must be provided.")
    domain_path = domain_file_path if domain_file_path is not None else working_directory_path / domain_file_name
    model_domain = partial_domain = DomainParser(domain_path=domain_path, partial_parsing=False).parse_domain()
    observations = []
    problem_files = list(test_set_dir_path.glob(f"{problem_prefix}*.pddl"))
    for test_problem_path in problem_files:
        trajectory_file_path = test_set_dir_path / f"{test_problem_path.stem}.trajectory"
        problem = ProblemParser(test_problem_path, partial_domain).parse_problem()
        observation = TrajectoryParser(partial_domain, problem).parse_trajectory(trajectory_file_path, executing_agents=executing_agents)
        observations.append(observation)

    if learning_algorithm in NUMERIC_ALGORITHMS:
        return NumericPerformanceCalculator(
            model_domain=model_domain,
            observations=observations,
            model_domain_path=domain_path,
            working_directory_path=working_directory_path,
            learning_algorithm=learning_algorithm,
        )

    elif learning_algorithm in MULTI_AGENT_ALGORITHMS:
        return MASamPerformanceCalculator(
            model_domain=model_domain,
            observations=observations,
            model_domain_path=domain_path,
            working_directory_path=working_directory_path,
            learning_algorithm=learning_algorithm,
        )

    return SemanticPerformanceCalculator(
        model_domain=model_domain,
        model_domain_path=domain_path,
        observations=observations,
        working_directory_path=working_directory_path,
        learning_algorithm=learning_algorithm,
    )



"""Module responsible for calculating our approach for numeric precision and recall."""
import csv
import logging
import uuid
from collections import defaultdict
from pathlib import Path
from typing import List, Dict, Tuple, Any, Union, Optional

from pddl_plus_parser.exporters import ProblemExporter
from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import (
    Domain,
    Observation,
    ActionCall,
    State,
    MultiAgentObservation,
    Operator,
    PDDLObject,
    Problem,
)

from sam_learning.core import VocabularyCreator
from utilities import LearningAlgorithmType, NegativePreconditionPolicy
from validators import run_validate_script, VALID_PLAN

SEMANTIC_PRECISION_STATS = [
    "action_name",
    "policy",
    "learning_algorithm",
    "num_trajectories",
    "precondition_precision",
    "precondition_recall",
    "effects_precision",
    "effects_recall",
]


def _calculate_precision_recall(
    num_false_negatives: Dict[str, int], num_false_positives: Dict[str, int], num_true_positives: Dict[str, int]
) -> Tuple[Dict[str, float], Dict[str, float]]:
    """Calculates the precision and recall values for each action.

    :param num_false_negatives: the number of false negatives for each action.
    :param num_false_positives: the number of false positives for each action.
    :param num_true_positives: the number of true positives for each action.
    :return: a tuple of two dictionaries, one for the precision values and one for the recall values.
    """
    precision_dict = defaultdict(float)
    recall_dict = defaultdict(float)
    for action_name, tp_rate in num_true_positives.items():
        if tp_rate == 0:
            precision_dict[action_name] = 1 if num_false_positives[action_name] == 0 else 0
            recall_dict[action_name] = 0
            continue

        precision_dict[action_name] = tp_rate / (tp_rate + num_false_positives[action_name])
        recall_dict[action_name] = tp_rate / (tp_rate + num_false_negatives[action_name])

    return precision_dict, recall_dict


class SemanticPerformanceCalculator:
    """Class responsible for calculating the semantic precision and recall of a model."""

    model_domain: Domain
    dataset_observations: List[Union[Observation, MultiAgentObservation]]
    learning_algorithm: LearningAlgorithmType
    combined_stats: List[Dict[str, Any]]
    logger: logging.Logger
    results_dir_path: Path
    _random_actions: List[List[ActionCall]]

    def __init__(
        self,
        model_domain: Domain,
        model_domain_path: Path,
        observations: List[Union[Observation, MultiAgentObservation]],
        working_directory_path: Path,
        learning_algorithm: LearningAlgorithmType,
    ):
        self.logger = logging.getLogger(__name__)
        self.model_domain = model_domain
        self.model_domain_path = model_domain_path
        self.dataset_observations = observations
        self.learning_algorithm = learning_algorithm
        self.combined_stats = []
        self.results_dir_path = working_directory_path / "results_directory"
        self.temp_dir_path = working_directory_path / "temp"
        self.temp_dir_path.mkdir(exist_ok=True)
        self.vocabulary_creator = VocabularyCreator()
        self._random_actions = []

    def _calculate_action_applicability_rate(
        self, action_call: ActionCall, learned_domain_path: Path, observed_state: State, problem_objects: Dict[str, PDDLObject],
    ) -> Tuple[int, int, int]:
        """Test whether an action is applicable in both the model domain and the generated domain.

        :param action_call: the action call that is tested for applicability.
        :param learned_domain_path: the domain that was learned using the action model learning algorithm.
        :param observed_state: the state that was observed in the trajectory data.
        :param problem_objects: the objects that were used in the problem definition.
        :return: a tuple containing the number of true positives, false positives and false negatives.
        """
        self.logger.debug(f"Calculating the applicability rate for the action - {action_call.name}")
        applicability_validation_problem = Problem(domain=self.model_domain)
        applicability_validation_problem.name = f"instance_{uuid.uuid4()}"
        applicability_validation_problem.objects = problem_objects
        applicability_validation_problem.initial_state_predicates = observed_state.state_predicates
        applicability_validation_problem.initial_state_fluents = observed_state.state_fluents
        self.logger.debug(
            f"Exporting a problem with the initial state as in {str(observed_state)} and no goals to validate whether the action is applicable."
        )
        current_problem_file_path = self.temp_dir_path / f"applicability_validation_problem_{uuid.uuid4()}.pddl"
        current_solution_file_path = self.temp_dir_path / f"applicability_validation_solution_{uuid.uuid4()}.solution"
        with open(current_solution_file_path, "wt") as solution_file:
            solution_file.write(str(action_call))

        ProblemExporter().export_problem(problem=applicability_validation_problem, export_path=current_problem_file_path)

        self.logger.debug(f"Exported the problem to {current_problem_file_path}, now validating the action's applicability.")
        applicable_in_model = self._calculate_applicability_in_state(current_problem_file_path, current_solution_file_path, self.model_domain_path)
        applicable_in_learned = self._calculate_applicability_in_state(current_problem_file_path, current_solution_file_path, learned_domain_path)

        current_problem_file_path.unlink()
        current_solution_file_path.unlink()
        return (
            int(applicable_in_learned == applicable_in_model and applicable_in_learned),
            int(applicable_in_learned and not applicable_in_model),
            int(not applicable_in_learned and applicable_in_model),
        )

    @staticmethod
    def _calculate_applicability_in_state(problem_file_path: Path, solution_file_path: Path, domain_file_path: Path) -> bool:
        """Calculate whether the action is applicable in the state.

        :param problem_file_path: the path to the problem file.
        :param solution_file_path: the path to the solution file.
        :param domain_file_path: the path to the domain file.
        :return: whether the action is applicable in the state.
        """
        validation_file_path = run_validate_script(
            domain_file_path=domain_file_path, problem_file_path=problem_file_path, solution_file_path=solution_file_path
        )
        with open(validation_file_path, "r", encoding="utf-8") as validation_file:
            validation_file_content = validation_file.read()

        validation_file_path.unlink()
        return VALID_PLAN in validation_file_content

    def _calculate_effects_difference_rate(
        self,
        observation: Observation,
        learned_domain: Domain,
        num_false_negatives: Dict[str, int],
        num_false_positives: Dict[str, int],
        num_true_positives: Dict[str, int],
    ) -> None:
        """Calculates the effects difference rate for each action.

        :param observation: the observation that is being tested.
        :param learned_domain: the domain that was learned using the action model learning algorithm.
        :param num_false_negatives: the dictionary mapping between the action name and the number of false negative
        :param num_false_positives: the dictionary mapping between the action name and the number of false positive
        :param num_true_positives: the dictionary mapping between the action name and the number of true positive
        """
        self.logger.info("Calculating effects difference rate for the observation.")
        for observation_triplet in observation.components:
            model_previous_state = observation_triplet.previous_state
            executed_action = observation_triplet.grounded_action_call
            model_next_state = observation_triplet.next_state
            if executed_action.name not in learned_domain.actions:
                continue

            try:
                learned_operator = Operator(
                    action=learned_domain.actions[executed_action.name],
                    domain=learned_domain,
                    grounded_action_call=executed_action.parameters,
                    problem_objects=observation.grounded_objects,
                )
                # cannot apply when the action is inapplicable since the linear regression is fitted on applicable numeric points.
                try:
                    learned_next_state = learned_operator.apply(model_previous_state)

                except ValueError:
                    self.logger.debug("The action is not applicable in the state.")
                    learned_next_state = model_previous_state.copy()
                    model_next_state = model_previous_state.copy()

                except ZeroDivisionError:
                    self.logger.debug("Tried to perform a division by zero!")
                    learned_next_state = model_previous_state.copy()
                    model_next_state = model_previous_state.copy()

                self.logger.debug("Validating if there are any false negatives.")
                model_state_predicates = {
                    predicate.untyped_representation
                    for grounded_predicate in model_next_state.state_predicates.values()
                    for predicate in grounded_predicate
                }
                learned_state_predicates = {
                    predicate.untyped_representation
                    for grounded_predicate in learned_next_state.state_predicates.values()
                    for predicate in grounded_predicate
                }
                if len(model_state_predicates) == 0 and len(learned_state_predicates) == 0:
                    num_true_positives[executed_action.name] += 1
                    continue

                num_true_positives[executed_action.name] += len(model_state_predicates.intersection(learned_state_predicates))
                num_false_positives[executed_action.name] += len(learned_state_predicates.difference(model_state_predicates))
                num_false_negatives[executed_action.name] += len(model_state_predicates.difference(learned_state_predicates))

            except ValueError:
                self.logger.debug("The action is not applicable in the state.")
                continue

    def calculate_preconditions_semantic_performance(
        self, learned_domain: Domain, learned_domain_path: Path
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Calculates the precision recall values of the learned preconditions.

        :param learned_domain: the action model that was learned using the action model learning algorithm
        :param learned_domain_path: the path to the learned domain.
        :return: the precision and recall dictionaries.
        """
        num_true_positives = defaultdict(int)
        num_false_negatives = defaultdict(int)
        num_false_positives = defaultdict(int)
        self.logger.debug("Starting to calculate the semantic preconditions performance")
        for index, observation in enumerate(self.dataset_observations):
            observation_objects = observation.grounded_objects
            for component in observation.components:
                action = component.grounded_action_call
                if action.name not in learned_domain.actions:
                    continue

                true_positive, false_positive, false_negative = self._calculate_action_applicability_rate(
                    action, learned_domain_path, component.previous_state, observation_objects,
                )
                num_true_positives[action.name] += true_positive
                num_false_positives[action.name] += false_positive
                num_false_negatives[action.name] += false_negative

        return _calculate_precision_recall(num_false_negatives, num_false_positives, num_true_positives)

    def calculate_effects_semantic_performance(self, learned_domain: Domain) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Calculates the precision recall values of the learned effects.

        :param learned_domain: the action model that was learned using the action model learning algorithm
        :return: the precision and recall dictionaries.
        """
        self.logger.info("Starting to calculate the semantic effects performance")
        num_true_positives = defaultdict(int)
        num_false_negatives = defaultdict(int)
        num_false_positives = defaultdict(int)
        if len(self.model_domain.predicates) == 0:
            return (
                {action_name: 1 for action_name in self.model_domain.actions.keys()},
                {action_name: 1 for action_name in self.model_domain.actions.keys()},
            )

        for observation in self.dataset_observations:
            self._calculate_effects_difference_rate(observation, learned_domain, num_false_negatives, num_false_positives, num_true_positives)

        return _calculate_precision_recall(num_false_negatives, num_false_positives, num_true_positives)

    def calculate_performance(self, learned_domain_path: Path, num_used_observations: int, policy: NegativePreconditionPolicy) -> None:
        """Calculate the semantic precision and recall of the learned domain.

        :param learned_domain_path: the path to the learned domain.
        :param num_used_observations: the number of observations used to learn the domain.
        :param policy: the policy to use for the negative preconditions.
        """
        learned_domain = DomainParser(domain_path=learned_domain_path, partial_parsing=False).parse_domain()
        self.logger.info("Starting to calculate the semantic preconditions performance of the learned domain.")
        preconditions_precision, preconditions_recall = self.calculate_preconditions_semantic_performance(learned_domain, learned_domain_path)
        self.logger.info("Starting to calculate the semantic effects performance of the learned domain.")
        effects_precision, effects_recall = self.calculate_effects_semantic_performance(learned_domain)
        for action_name in self.model_domain.actions:
            action_stats = {
                "learning_algorithm": self.learning_algorithm.name,
                "policy": policy.name,
                "action_name": action_name,
                "num_trajectories": num_used_observations,
                "precondition_precision": preconditions_precision.get(action_name, 1),
                "precondition_recall": preconditions_recall.get(action_name, 0),
                "effects_precision": effects_precision.get(action_name, 1),
                "effects_recall": effects_recall.get(action_name, 1),
            }
            self.combined_stats.append(action_stats)

    def export_semantic_performance(self, fold_num: int, iteration_num: Optional[int] = None) -> None:
        """Exports the precision values of the learned preconditions to a CSV file.

        :param fold_num: the fold number.
        :param iteration_num: the iteration number.
        """
        iterations_suffix = f"{iteration_num}" if iteration_num is not None else ""
        statistics_path = (
            self.results_dir_path / f"{self.learning_algorithm.name}_{self.model_domain.name}_{fold_num}_{iterations_suffix}_semantic_performance.csv"
        )
        with open(statistics_path, "wt", newline="") as statistics_file:
            stats_writer = csv.DictWriter(statistics_file, fieldnames=SEMANTIC_PRECISION_STATS)
            stats_writer.writeheader()
            for data_line in self.combined_stats:
                stats_writer.writerow(data_line)

    def export_combined_semantic_performance(self) -> None:
        """Export the numeric learning statistics to a CSV report file."""
        statistics_path = self.results_dir_path / f"{self.learning_algorithm.name}_{self.model_domain.name}_combined_semantic_performance.csv"
        with open(statistics_path, "wt", newline="") as statistics_file:
            stats_writer = csv.DictWriter(statistics_file, fieldnames=SEMANTIC_PRECISION_STATS)
            stats_writer.writeheader()
            for data_line in self.combined_stats:
                stats_writer.writerow(data_line)



import logging
import os
import signal
import subprocess
import sys
from pathlib import Path
from typing import Dict

from pddl_plus_parser.exporters import MetricFFParser

METRIC_FF_DIRECTORY = os.environ["METRIC_FF_DIRECTORY"]

MAX_RUNNING_TIME = 5  # seconds


class MetricFFSolver:
    """Class designated to use to activate the metric-FF solver on the cluster and parse its result."""

    logger: logging.Logger

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.parser = MetricFFParser()

    @staticmethod
    def _extract_solver_error(solution_path: Path) -> str:
        """Extracts the error from the solution file.

        :param solution_path: the path to the solution file.
        :return: the error message.
        """
        with open(solution_path, "rb") as solution_file:
            solution_content = solution_file.read()
            return solution_content.decode("utf-8", errors="ignore")

    def _run_metric_ff_process(
        self, run_command: str, solution_path: Path, problem_file_path: Path, solving_stats: Dict[str, str], solving_timeout: int = MAX_RUNNING_TIME,
    ) -> None:
        """Runs the metric-ff process."""
        self.logger.info(f"Metric-FF solver is working on - {problem_file_path.stem}")
        process = subprocess.Popen(run_command, shell=True)
        try:
            process.wait(timeout=solving_timeout)
        except subprocess.TimeoutExpired:
            self.logger.warning(f"Metric-FF solver took more than {solving_timeout} seconds to finish.")
            os.kill(process.pid, signal.SIGTERM)
            os.system("pkill -f ./ff")
            solution_path.unlink(missing_ok=True)
            solving_stats[problem_file_path.stem] = "timeout"
            return

        if process.returncode is None:
            solving_stats[problem_file_path.stem] = "timeout"
            solution_path.unlink(missing_ok=True)
            return

        if process.returncode != 0:
            self.logger.warning(f"Metric FF Solver returned status code {process.returncode}.")
            if not solution_path.exists():
                solving_stats[problem_file_path.stem] = "solver_error"
                return

            solving_status = self.parser.get_solving_status(solution_path)[0]
            if solving_status != "ok" and solving_status != "no-solution":
                solving_stats[problem_file_path.stem] = "solver_error"
                error = self._extract_solver_error(solution_path)
                self.logger.warning(f"Metric FF Solver encountered an error - {problem_file_path.stem}")
                self.logger.warning(error)
                solution_path.unlink(missing_ok=True)
                return

        self.logger.info("Metric FF Solver finished its execution!")
        solving_status = self.parser.get_solving_status(solution_path)[0]
        if solving_status == "ok":
            self.logger.info(f"Solver succeeded in solving problem - {problem_file_path.stem}")
            solving_stats[problem_file_path.stem] = solving_status
            self.parser.parse_plan(solution_path, solution_path)

        elif solving_status == "no-solution":
            self.logger.warning(f"Metric FF Solver could not solve problem - {problem_file_path.stem}")
            solving_stats[problem_file_path.stem] = "no_solution"
            solution_path.unlink(missing_ok=True)

    def solve_problem(
        self,
        domain_file_path: Path,
        problem_file_path: Path,
        problems_directory_path: Path,
        solving_stats: Dict[str, str],
        solving_timeout: int,
        tolerance: float,
    ) -> None:
        """Solves a single problem using the Metric FF algorithm.

        :param domain_file_path: the path to the domain file.
        :param problem_file_path: the path to the problem file.
        :param problems_directory_path: the path to the problems' directory.
        :param solving_stats: the statistics of the solving process.
        :param solving_timeout: the timeout for the solver.
        :param tolerance: the numeric tolerance to use.
        """
        os.chdir(METRIC_FF_DIRECTORY)
        self.logger.debug(f"Starting to work on solving problem - {problem_file_path.stem}")
        solution_path = problems_directory_path / f"{problem_file_path.stem}.solution"
        run_command = f"./ff -o {domain_file_path} -f {problem_file_path} -s 0 -t {tolerance} > {solution_path}"
        self._run_metric_ff_process(run_command, solution_path, problem_file_path, solving_stats, solving_timeout)

    def execute_solver(
        self,
        problems_directory_path: Path,
        domain_file_path: Path,
        solving_timeout: int = MAX_RUNNING_TIME,
        problems_prefix: str = "pfile",
        tolerance: float = 0.1,
    ) -> Dict[str, str]:
        """Solves numeric and PDDL+ problems using the Metric-FF algorithm and outputs the solution into a file.

        :param problems_directory_path: the path to the problems directory.
        :param domain_file_path: the path to the domain file.
        :param solving_timeout: the timeout for the solver.
        :param problems_prefix: the prefix of the problems files.
        :param tolerance: the numeric tolerance for errors in the Metric-FF solver.
        """
        solving_stats = {}
        self.logger.info("Starting to solve the input problems using Metic-FF solver.")
        for problem_file_path in problems_directory_path.glob(f"{problems_prefix}*.pddl"):
            self.solve_problem(domain_file_path, problem_file_path, problems_directory_path, solving_stats, solving_timeout, tolerance)

        return solving_stats


if __name__ == "__main__":
    args = sys.argv
    logging.basicConfig(format="%(asctime)s %(name)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
    solver = MetricFFSolver()
    solver.execute_solver(
        problems_directory_path=Path(args[1]), domain_file_path=Path(args[2]), problems_prefix=args[3], solving_timeout=int(args[4])
    )



"""Module responsible for calculating our approach for numeric precision and recall."""
import csv
import os
from collections import defaultdict
from pathlib import Path
from typing import List, Dict, Union, Optional

import sklearn
from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import Domain, Observation, MultiAgentObservation

from statistics.performance_calculation_utils import _ground_executed_action
from statistics.semantic_performance_calculator import SemanticPerformanceCalculator
from utilities import LearningAlgorithmType
from utilities.util_types import NUMERIC_PRECISION, NegativePreconditionPolicy

NUMERIC_PERFORMANCE_STATS = [
    "learning_algorithm",
    "policy",
    "action_name",
    "epsilon_precision",
    "num_trajectories",
    "precondition_precision",
    "precondition_recall",
    "effects_precision",
    "effects_recall",
    "effects_mse",
]

MAX_INT = 10 ** 6


class NumericPerformanceCalculator(SemanticPerformanceCalculator):
    """Class responsible for calculating the numeric precision and recall of a model."""

    def __init__(
        self,
        model_domain: Domain,
        model_domain_path: Path,
        observations: List[Union[Observation, MultiAgentObservation]],
        working_directory_path: Path,
        learning_algorithm: LearningAlgorithmType,
    ):
        super().__init__(model_domain, model_domain_path, observations, working_directory_path, learning_algorithm)
        self.epsilon_precision = 10 ** (-int(os.environ.get(NUMERIC_PRECISION, 4)))

    def calculate_effects_mse(self, learned_domain: Domain) -> Dict[str, float]:
        """Calculates the effects MSE value using the actual state fluents and the ones generated using the learned
            action.

        Note:
            The final value is averaged over the number of states per action.

        :param learned_domain: the domain that was learned by the action model learning algorithm.
        :return: a mapping between the action name and its MSE value.
        """
        squared_errors = defaultdict(list)
        mse_values = {action_name: 0 for action_name in learned_domain.actions.keys()}
        for observation in self.dataset_observations:
            for observation_component in observation.components:
                action_call = observation_component.grounded_action_call
                previous_state = observation_component.previous_state
                if action_call.name not in learned_domain.actions:
                    continue

                grounded_operator = _ground_executed_action(action_call, learned_domain, observation.grounded_objects)
                grounded_model_operator = _ground_executed_action(action_call, self.model_domain, observation.grounded_objects)
                try:
                    next_state = grounded_operator.apply(previous_state, allow_inapplicable_actions=False)
                    model_next_state = grounded_model_operator.apply(previous_state, allow_inapplicable_actions=False)

                except ValueError:
                    self.logger.debug("The action is not applicable in the state.")
                    next_state = previous_state.copy()
                    # since the learned action is not applicable in the state there is no point to compare
                    # with an action that is not applicable in the model domain.
                    model_next_state = previous_state.copy()

                except ZeroDivisionError:
                    # If the action performed division by zero the MSE is set to the maximum value.
                    squared_errors[action_call.name].append(MAX_INT)
                    continue

                values = [
                    (next_state.state_fluents[fluent].value, model_next_state.state_fluents[fluent].value)
                    for fluent in next_state.state_fluents.keys()
                ]
                actual_values, expected_values = zip(*values)
                state_mse = sklearn.metrics.mean_squared_error(expected_values, actual_values)
                squared_errors[action_call.name].append(state_mse)

        mse_values.update({action_name: sum(square_errors) / len(square_errors) for action_name, square_errors in squared_errors.items()})
        return mse_values

    def calculate_performance(self, learned_domain_path: Path, num_used_observations: int, policy: NegativePreconditionPolicy) -> None:
        """Calculates the model's performance with both the precision and the recall values calculated.

        :param learned_domain_path: the path to the learned action model.
        :param num_used_observations: the number of observations used to learn the action model.
        :param policy: the policy to use for the negative preconditions.
        """
        super().calculate_performance(learned_domain_path, num_used_observations, policy)
        learned_domain = DomainParser(domain_path=learned_domain_path, partial_parsing=False).parse_domain()
        effects_mse = self.calculate_effects_mse(learned_domain)
        for action in self.model_domain.actions:
            for action_stats in self.combined_stats:
                if action_stats["action_name"] == action and action_stats["num_trajectories"] == num_used_observations:
                    action_stats["effects_mse"] = effects_mse.get(action, 0)
                    action_stats["epsilon_precision"] = self.epsilon_precision
                    break

        self.logger.info(f"Finished calculating the numeric learning performance for {self.learning_algorithm.name}.")

    def export_semantic_performance(self, fold_num: int, iteration_num: Optional[int] = None) -> None:
        """Export the numeric learning statistics to a CSV report file.

        :param fold_num: the fold number of the current experiment.
        :param iteration_num: the iteration number of the current experiment.
        """
        iteration_suffix = f"_{iteration_num}" if iteration_num is not None else ""
        statistics_path = (
            self.results_dir_path / f"{self.learning_algorithm.name}_{self.model_domain.name}"
            f"_numeric_learning_performance_stats_fold_{fold_num}{iteration_suffix}.csv"
        )
        with open(statistics_path, "wt", newline="") as statistics_file:
            stats_writer = csv.DictWriter(statistics_file, fieldnames=NUMERIC_PERFORMANCE_STATS)
            stats_writer.writeheader()
            for data_line in self.combined_stats:
                stats_writer.writerow(data_line)

    def export_combined_semantic_performance(self) -> None:
        """Export the numeric learning statistics to a CSV report file."""
        statistics_path = self.results_dir_path / f"{self.learning_algorithm.name}_{self.model_domain.name}" "combined_numeric_performance.csv"
        with open(statistics_path, "wt", newline="") as statistics_file:
            stats_writer = csv.DictWriter(statistics_file, fieldnames=NUMERIC_PERFORMANCE_STATS)
            stats_writer.writeheader()
            for data_line in self.combined_stats:
                stats_writer.writerow(data_line)



"""The POL main framework - Compile, Learn and Plan."""
import argparse
import logging
import os
import shutil
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import List, Optional, Dict, Tuple, Any, Union

from pddl_plus_parser.lisp_parsers import DomainParser, TrajectoryParser, ProblemParser
from pddl_plus_parser.models import Observation, Domain, MultiAgentObservation

from experiments.experiments_consts import MAX_SIZE_MB, DEFAULT_SPLIT, DEFAULT_NUMERIC_TOLERANCE, NUMERIC_ALGORITHMS
from sam_learning.core import LearnerDomain
from statistics.learning_statistics_manager import LearningStatisticsManager
from statistics.utils import init_semantic_performance_calculator
from utilities import LearningAlgorithmType, SolverType, NegativePreconditionPolicy
from utilities.k_fold_split import KFoldSplit
from validators import DomainValidator

PLANNER_EXECUTION_TIMEOUT = 1800  # 30 minutes
NUM_TRIPLETS_PER_TESTING = 10


def configure_iteration_logger(args: argparse.Namespace):
    """Configures the logger for the numeric action model learning algorithms evaluation experiments."""
    iteration_number = int(args.iteration_number) if "iteration_number" in args else "all"
    learning_algorithm = LearningAlgorithmType(args.learning_algorithm)
    local_logs_parent_path = os.environ.get("LOCAL_LOGS_PATH", args.working_directory_path)
    working_directory_path = Path(local_logs_parent_path)
    logs_directory_path = working_directory_path / "logs"
    try:
        logs_directory_path.mkdir(exist_ok=True)
    except PermissionError:
        # This is a hack to not fail and just avoid logging in case the directory cannot be created
        return

    # Create a rotating file handler
    max_bytes = MAX_SIZE_MB * 1024 * 1024  # Convert megabytes to bytes
    file_handler = RotatingFileHandler(
        logs_directory_path / f"log_{args.domain_file_name}_fold_{learning_algorithm.name}_{args.fold_number}_{iteration_number}.log",
        maxBytes=max_bytes,
        backupCount=1,
    )
    stream_handler = logging.StreamHandler()

    # Create a formatter and set it for the handler
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    file_handler.setFormatter(formatter)
    if args.debug:
        logging.basicConfig(datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO, handlers=[file_handler, stream_handler])
    else:
        logging.basicConfig(datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO, handlers=[file_handler])


class ParallelExperimentRunner:
    """Class that represents the POL framework."""

    logger: logging.Logger
    working_directory_path: Path
    k_fold: KFoldSplit
    domain_file_name: str
    learning_statistics_manager: LearningStatisticsManager
    _learning_algorithm: LearningAlgorithmType
    domain_validator: DomainValidator
    fluents_map: Dict[str, List[str]]

    def __init__(
        self,
        working_directory_path: Path,
        domain_file_name: str,
        learning_algorithm: LearningAlgorithmType,
        problem_prefix: str = "pfile",
        running_triplets_experiment: bool = False,
        executing_agents: List[str] = None,
    ):
        self.logger = logging.getLogger(__name__)
        self.working_directory_path = working_directory_path
        self.k_fold = KFoldSplit(working_directory_path=working_directory_path, domain_file_name=domain_file_name, n_split=DEFAULT_SPLIT)
        self.domain_file_name = domain_file_name
        self._learning_algorithm = learning_algorithm
        self.semantic_performance_calc = None
        self.domain_validator = DomainValidator(
            self.working_directory_path, learning_algorithm, self.working_directory_path / domain_file_name, problem_prefix=problem_prefix,
        )
        self.problem_prefix = problem_prefix
        self.executing_agents = executing_agents
        self._running_triplets_experiment = running_triplets_experiment
        self.learning_statistics_manager = LearningStatisticsManager(
            working_directory_path=working_directory_path,
            domain_path=self.working_directory_path / domain_file_name,
            learning_algorithm=learning_algorithm,
        )

    def _init_semantic_performance_calculator(self, fold_num: int) -> None:
        """Initializes the algorithm of the semantic precision - recall calculator."""
        self.semantic_performance_calc = init_semantic_performance_calculator(
            working_directory_path=self.working_directory_path,
            domain_file_name=self.domain_file_name,
            learning_algorithm=self._learning_algorithm,
            test_set_dir_path=self.working_directory_path / "performance_evaluation_trajectories" / f"fold_{fold_num}",
            problem_prefix=self.problem_prefix,
            executing_agents=self.executing_agents,
        )

    def _apply_learning_algorithm(
        self, partial_domain: Domain, allowed_observations: List[Observation], test_set_dir_path: Path
    ) -> Tuple[LearnerDomain, Dict[str, Any]]:
        raise NotImplementedError

    def _export_learned_domain(self, learned_domain: LearnerDomain, test_set_path: Path, file_name: Optional[str] = None) -> Path:
        """Exports the learned domain into a file so that it will be used to solve the test set problems.

        :param learned_domain: the domain that was learned by the action model learning algorithm.
        :param test_set_path: the path to the test set directory where the domain would be exported to.
        :param file_name: the name of the file to export the domain to.
        """
        domain_file_name = file_name or self.domain_file_name
        domain_path = test_set_path / domain_file_name
        with open(domain_path, "wt") as domain_file:
            domain_file.write(learned_domain.to_pddl(should_simplify=False))

        return domain_path

    def create_transitions_based_training_set(
        self, complete_train_set: Union[List[MultiAgentObservation], List[Observation]], num_triplets_per_testing: int = NUM_TRIPLETS_PER_TESTING
    ) -> Union[List[MultiAgentObservation], List[Observation]]:
        """Creates new observations containing only part of the transitions for the learning process.

        :param complete_train_set: the complete training set containing all the observations.
        :param num_triplets_per_testing: the number of triplets to use in each observation.
        :return: the new training set containing only part of the transitions.
        """
        transition_training_set = []
        self.logger.debug(f"Extracting {num_triplets_per_testing} sized observations from each trajectory.")
        for observation in complete_train_set:
            if len(observation) % num_triplets_per_testing != 0:
                self.logger.debug(f"Will remove some of the transitions to make the number of transitions divisible by {num_triplets_per_testing}.")

            for index in range(0, len(observation), num_triplets_per_testing):
                if len(observation.components[index : index + num_triplets_per_testing]) < num_triplets_per_testing:
                    break

                new_obs = MultiAgentObservation(executing_agents=self.executing_agents) if self.executing_agents is not None else Observation()
                new_obs.add_problem_objects(observation.grounded_objects)
                new_obs.components = observation.components[index : index + num_triplets_per_testing]
                transition_training_set.append(new_obs)

        return transition_training_set

    def _export_domain_and_backup(
        self,
        allowed_observations: Union[List[Observation], List[MultiAgentObservation]],
        fold_number: int,
        learned_model: LearnerDomain,
        negative_preconditions_policy: NegativePreconditionPolicy,
        test_set_dir_path: Path,
    ) -> Path:
        """Exports the domain to the test set directory and another backup to the results directory.

        :param allowed_observations: the observations that were used in the learning process.
        :param fold_number: the number of the fold that is currently running.
        :param learned_model: the domain learned by the action model learning algorithm.
        :param negative_preconditions_policy: the policy to use for the negative preconditions.
        :param test_set_dir_path: the path to the test set directory.
        :return: the path of the PDDL domain in the test set directory.
        """
        domain_file_path = self._export_learned_domain(learned_model, test_set_dir_path)
        domains_backup_dir_path = self.working_directory_path / "results_directory" / "domains_backup"
        domains_backup_dir_path.mkdir(exist_ok=True)
        experiment_type_extension = (
            f"{sum([len(obs) for obs in allowed_observations])}_triplets"
            if self._running_triplets_experiment
            else f"{len(allowed_observations)}_trajectories"
        )
        shutil.copy(
            domain_file_path,
            domains_backup_dir_path / f"{self._learning_algorithm.name}_fold_{fold_number}_{learned_model.name}"
            f"_{experiment_type_extension}_{negative_preconditions_policy.name}_policy.pddl",
        )
        return domain_file_path

    def read_domain_file(self, train_set_dir_path: Path) -> Domain:
        """Reads the domain file and returns the domain object.

        :param train_set_dir_path: the path of the directory containing the domain file.
        :return: the parsed domain object.
        """
        partial_domain_path = train_set_dir_path / self.domain_file_name
        return DomainParser(domain_path=partial_domain_path, partial_parsing=True).parse_domain()

    def collect_observations(self, train_set_dir_path: Path, partial_domain: Domain) -> Union[List[Observation], List[MultiAgentObservation]]:
        """Collects all the observations from the trajectories in the train set directory.

        :param train_set_dir_path: the path to the directory containing the trajectories.
        :param partial_domain: the partial domain without the actions' preconditions and effects.
        :return: the allowed observations.
        """
        allowed_observations = []
        sorted_trajectory_paths = sorted(train_set_dir_path.glob("*.trajectory"))  # for consistency
        for index, trajectory_file_path in enumerate(sorted_trajectory_paths):
            # assuming that the folders were created so that each folder contains only the correct number of trajectories, i.e., iteration_number
            problem_path = train_set_dir_path / f"{trajectory_file_path.stem}.pddl"
            problem = ProblemParser(problem_path, partial_domain).parse_problem()
            complete_observation = TrajectoryParser(partial_domain, problem).parse_trajectory(trajectory_file_path, self.executing_agents)
            allowed_observations.append(complete_observation)

        return allowed_observations

    def learn_model_offline(
        self,
        fold_num: int,
        train_set_dir_path: Path,
        test_set_dir_path: Path,
        negative_preconditions_policy: NegativePreconditionPolicy = NegativePreconditionPolicy.no_remove,
    ) -> None:
        """Learns the model of the environment by learning from the input trajectories.

        :param fold_num: the index of the current folder that is currently running.
        :param train_set_dir_path: the directory containing the trajectories in which the algorithm is learning from.
        :param test_set_dir_path: the directory containing the test set problems in which the learned model should be
            used to solve.
        :param negative_preconditions_policy: the policy to use for the negative preconditions.
        """
        self.logger.info(f"Starting the learning phase for the fold - {fold_num}!")
        partial_domain_path = train_set_dir_path / self.domain_file_name
        partial_domain = DomainParser(domain_path=partial_domain_path, partial_parsing=True).parse_domain()
        allowed_observations = self.collect_observations(train_set_dir_path, partial_domain)
        self.logger.info(f"Learning the action model using {len(allowed_observations)} trajectories!")
        learned_model, learning_report = self._apply_learning_algorithm(partial_domain, allowed_observations, test_set_dir_path)
        self.learning_statistics_manager.add_to_action_stats(
            allowed_observations, learned_model, learning_report, policy=negative_preconditions_policy
        )
        learned_domain_path = self.validate_learned_domain(
            allowed_observations, learned_model, test_set_dir_path, fold_num, learning_report["learning_time"]
        )
        self.semantic_performance_calc.calculate_performance(learned_domain_path, len(allowed_observations), policy=negative_preconditions_policy)

    def run_fold_iteration(self, fold_num: int, train_set_dir_path: Path, test_set_dir_path: Path, iteration_number: int) -> None:
        """Runs the numeric action model learning algorithms on the input fold.

        :param fold_num: the number of the fold to run.
        :param train_set_dir_path: the path to the directory containing the training set problems.
        :param test_set_dir_path: the path to the directory containing the test set problems.
        :param iteration_number: the current iteration number.
        """
        self.logger.info(f"Running fold {fold_num} iteration {iteration_number}")
        self._init_semantic_performance_calculator(fold_num)
        self.learn_model_offline(fold_num, train_set_dir_path, test_set_dir_path)
        self.domain_validator.write_statistics(fold_num, iteration_number)
        self.semantic_performance_calc.export_semantic_performance(fold_num, iteration_number)
        self.learning_statistics_manager.export_action_learning_statistics(fold_number=fold_num, iteration_num=iteration_number)

    def validate_learned_domain(
        self,
        allowed_observations: Union[List[Observation], List[MultiAgentObservation]],
        learned_model: LearnerDomain,
        test_set_dir_path: Path,
        fold_number: int,
        learning_time: float,
        negative_preconditions_policy: NegativePreconditionPolicy = NegativePreconditionPolicy.no_remove,
    ) -> Path:
        """Validates that using the learned domain both the used and the test set problems can be solved.

        :param allowed_observations: the observations that were used in the learning process.
        :param learned_model: the domain that was learned using POL.
        :param test_set_dir_path: the path to the directory containing the test set problems.
        :param fold_number: the number of the fold that is currently running.
        :param learning_time: the time it took to learn the domain (in seconds).
        :param negative_preconditions_policy: the policy to use for the negative preconditions.
        :return: the path for the learned domain.
        """
        domain_file_path = self._export_domain_and_backup(
            allowed_observations, fold_number, learned_model, negative_preconditions_policy, test_set_dir_path
        )
        self.logger.debug("Checking that the test set problems can be solved using the learned domain.")
        portfolio = [SolverType.metric_ff, SolverType.enhsp] if self._learning_algorithm in NUMERIC_ALGORITHMS else [SolverType.fast_downward]
        if "ABORT_SOLVING" in os.environ:
            self.logger.debug("Skipping the validation of the learned domain due to request in the experiment.")
            return domain_file_path

        self.domain_validator.validate_domain(
            tested_domain_file_path=domain_file_path,
            test_set_directory_path=test_set_dir_path,
            used_observations=allowed_observations,
            tolerance=DEFAULT_NUMERIC_TOLERANCE,
            timeout=os.environ.get("PLANNER_EXECUTION_TIMEOUT", PLANNER_EXECUTION_TIMEOUT),
            learning_time=learning_time,
            solvers_portfolio=portfolio,
            preconditions_removal_policy=negative_preconditions_policy,
        )

        return domain_file_path



import random
from unittest.mock import patch

import numpy as np
import pytest
from pandas import DataFrame
from pddl_plus_parser.models import PDDLFunction, Precondition, NumericalExpressionTree
from scipy.spatial import ConvexHull

from sam_learning.core.numeric_learning.convex_hull_learner import ConvexHullLearner
from sam_learning.core.numeric_learning.numeric_utils import remove_complex_linear_dependencies, construct_pddl_inequality_scheme

TEST_ACTION_NAME = "test_action"

random.seed(42)


@pytest.fixture
def convex_hull_learner() -> ConvexHullLearner:
    domain_functions = {
        "x": PDDLFunction(name="x", signature={}),
        "y": PDDLFunction(name="y", signature={}),
        "z": PDDLFunction(name="z", signature={}),
        "w": PDDLFunction(name="w", signature={}),
    }
    return ConvexHullLearner(TEST_ACTION_NAME, domain_functions)


def test_construct_pddl_inequality_scheme_with_simple_2d_four_equations_returns_correct_representation(convex_hull_learner: ConvexHullLearner,):
    np.random.seed(42)
    left_side_coefficients = np.random.randint(10, size=(4, 2))
    right_side_points = np.random.randint(10, size=4)

    inequalities = construct_pddl_inequality_scheme(left_side_coefficients, right_side_points, ["(x)", "(y)"])
    assert len(inequalities) == 4
    for ineq in inequalities:
        assert ineq.startswith("(<=")


def test_create_disjunctive_preconditions_with_input_dataframe_of_size_one_returns_a_single_precondition_with_and_condition(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {"(x)": [2], "(y)": [3], "(z)": [-1]}
    dataframe = DataFrame(pre_state_data)
    equality_condition = convex_hull_learner._create_disjunctive_preconditions(dataframe, equality_conditions=[])
    assert len(equality_condition.operands) == 3
    assert equality_condition.binary_operator == "and"
    for operand in equality_condition.operands:
        assert operand.to_pddl().startswith("(=")


def test_create_disjunctive_preconditions_with_input_dataframe_of_size_two_returns_a_single_precondition_with_or_condition(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {"(x)": [2, 18], "(y)": [3, 32], "(z)": [-1, 17]}
    dataframe = DataFrame(pre_state_data)
    equality_condition = convex_hull_learner._create_disjunctive_preconditions(dataframe, equality_conditions=[])
    assert len(equality_condition.operands) == 2
    assert equality_condition.binary_operator == "or"
    for operand in equality_condition.operands:
        assert isinstance(operand, Precondition)
        assert operand.binary_operator == "and"
        assert len(operand.operands) == 3
        for sub_operand in operand.operands:
            assert sub_operand.to_pddl().startswith("(=")


def test_create_disjunctive_preconditions_with_input_dataframe_of_size_one_and_equality_condition_creates_combined_condition_correctly(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {"(x)": [20], "(y)": [39], "(z)": [-12]}
    test_equality_conditions = ["(= (x ) 2)", "(= (y ) 3)", "(= (z ) -1)"]
    dataframe = DataFrame(pre_state_data)
    equality_condition = convex_hull_learner._create_disjunctive_preconditions(dataframe, equality_conditions=test_equality_conditions)
    assert len(equality_condition.operands) == 2 * len(test_equality_conditions)
    assert equality_condition.binary_operator == "and"
    for operand in equality_condition.operands:
        assert operand.to_pddl().startswith("(=")


def test_create_disjunctive_preconditions_with_input_dataframe_of_size_two_and_equality_condition_creates_combined_condition_correctly(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {"(x)": [2, 18], "(y)": [3, 32], "(z)": [-1, 17]}
    dataframe = DataFrame(pre_state_data)
    test_equality_conditions = ["(= (x ) 97)", "(= (y ) 21)"]
    result_condition = convex_hull_learner._create_disjunctive_preconditions(dataframe, equality_conditions=test_equality_conditions)
    assert len(result_condition.operands) == 3
    assert len([op for op in result_condition.operands if isinstance(op, NumericalExpressionTree)]) == 2
    assert len([op for op in result_condition.operands if isinstance(op, Precondition)]) == 1


def test_construct_single_dimension_inequalities_with_input_dataframe_of_size_one_returns_correct_inequalities(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {"(x)": [2, 18], "(y)": [3, 32], "(z)": [-1, 17]}
    dataframe = DataFrame(pre_state_data)
    inequality_precondition = convex_hull_learner._construct_single_dimension_inequalities(dataframe["(x)"])
    assert inequality_precondition.binary_operator == "and"
    assert len(inequality_precondition.operands) == 2
    inequalities = [op.to_pddl() for op in inequality_precondition.operands]
    assert "(>= (x ) 2)" in inequalities
    assert "(<= (x ) 18)" in inequalities


def test_construct_single_dimension_inequalities_with_input_lower_bound_equal_to_upper_bound_returns_correct_condition(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {"(x)": [2, 2], "(y)": [3, 32], "(z)": [-1, 17]}
    dataframe = DataFrame(pre_state_data)
    inequality_precondition = convex_hull_learner._construct_single_dimension_inequalities(dataframe["(x)"])
    assert inequality_precondition.binary_operator == "and"
    assert len(inequality_precondition.operands) == 1
    equality = inequality_precondition.operands.pop()
    assert "(= (x ) 2)" == equality.to_pddl()


def test_construct_safe_linear_inequalities_when_the_number_of_samples_is_one_creates_a_single_condition(convex_hull_learner: ConvexHullLearner,):
    convex_hull_learner.add_new_point(
        {"(x )": 2, "(y )": 3, "(z )": -1,}
    )
    inequality_precondition = convex_hull_learner.construct_safe_linear_inequalities(None)
    assert inequality_precondition.binary_operator == "and"
    assert len(inequality_precondition.operands) == 3
    assert {op.to_pddl() for op in inequality_precondition.operands} == {
        "(= (x ) 2)",
        "(= (y ) 3)",
        "(= (z ) -1)",
    }


def test_construct_safe_linear_inequalities_when_the_number_of_samples_is_two_creates_a_two_conditions_in_or_statement(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {"(x )": [2, 3], "(y )": [3, 5], "(z )": [-1, 7]}
    for i in range(2):
        convex_hull_learner.add_new_point(
            {"(x )": pre_state_data["(x )"][i], "(y )": pre_state_data["(y )"][i], "(z )": pre_state_data["(z )"][i],}
        )
    inequality_precondition = convex_hull_learner.construct_safe_linear_inequalities(None)
    print(str(inequality_precondition))
    assert inequality_precondition.binary_operator == "and"
    assert len(inequality_precondition.operands) == 2 + 2  # 2 for the CH and 2 for complementing the base
    assert all([isinstance(op, NumericalExpressionTree) for op in inequality_precondition.operands])
    assert len([cond.to_pddl() for cond in inequality_precondition.operands if cond.to_pddl().startswith("(=")]) == 2
    assert len([cond.to_pddl() for cond in inequality_precondition.operands if cond.to_pddl().startswith("(<=")]) == 2


def test_construct_safe_linear_inequalities_when_the_number_of_is_smaller_than_needed_dimensions_but_larger_than_two_creates_a_convex_hull_from_the_samples(
    convex_hull_learner: ConvexHullLearner,
):

    for i in range(3):
        convex_hull_learner.add_new_point(
            {"(x )": random.randint(0, 100), "(y )": random.randint(0, 100), "(z )": random.randint(0, 100),}
        )
    inequality_precondition = convex_hull_learner.construct_safe_linear_inequalities(None)
    assert inequality_precondition.binary_operator == "and"
    assert len(inequality_precondition.operands) == 3 + 1  # 3 for the CH and 1 for complementing the base
    assert all([isinstance(op, NumericalExpressionTree) for op in inequality_precondition.operands])
    print(str(inequality_precondition))


def test_create_convex_hull_linear_inequalities_returns_correct_conditions_when_constant_is_observed(convex_hull_learner: ConvexHullLearner,):
    pre_state_data = {"(x)": [1, 0, 0], "(y)": [0, 1, 0], "(z)": [0, 0, 0]}
    pre_state_df = DataFrame(pre_state_data)
    (coefficients, border_point, transformed_vars, span_verification_conditions,) = convex_hull_learner._create_convex_hull_linear_inequalities(
        pre_state_df, display_mode=False
    )
    assert span_verification_conditions == ["(= (z) 0.0)"]
    print(span_verification_conditions)


def test_create_convex_hull_linear_inequalities_returns_correct_conditions_when_constant_is_observed_and_there_is_enough_data_to_create_convex_hull(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {"(x)": [1, 0, 0, 1], "(y)": [0, 1, 0, 1], "(z)": [0, 0, 1, 1]}
    pre_state_df = DataFrame(pre_state_data)
    (coefficients, border_point, transformed_vars, span_verification_conditions,) = convex_hull_learner._create_convex_hull_linear_inequalities(
        pre_state_df, display_mode=False
    )
    assert span_verification_conditions == []


def test_create_convex_hull_linear_inequalities_returns_correct_conditions_when_when_remaining_only_one_dimension_returns_correct_values(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {
        "(x)": [2, 1, 3, 4],
        "(y)": [1, 1, 1, 1],
    }
    pre_state_df = DataFrame(pre_state_data)
    (coefficients, border_point, transformed_vars, span_verification_conditions,) = convex_hull_learner._create_convex_hull_linear_inequalities(
        pre_state_df, display_mode=False
    )
    assert span_verification_conditions == ["(= (- (y) 1) 0.0)"]
    assert coefficients == [[-1], [1]]
    assert transformed_vars == ["(* (- (x) 2) -1)"]
    assert border_point == [2, 1]


def test_create_convex_hull_linear_inequalities_when_for_some_reason_the_gs_returns_a_base_with_larger_dimension_than_original_points_just_uses_original_points(
    convex_hull_learner: ConvexHullLearner,
):
    with patch("sam_learning.core.numeric_learning.numeric_utils.extended_gram_schmidt") as mock_gs:
        mock_gs.return_value = [[1, 0], [0, 1], [0, 0], [0, 23]]

        pre_state_data = {
            "(x)": [0, -1, 0],
            "(y)": [0, 0, -1],
        }
        pre_state_df = DataFrame(pre_state_data)
        (coefficients, border_point, transformed_vars, span_verification_conditions,) = convex_hull_learner._create_convex_hull_linear_inequalities(
            pre_state_df, display_mode=False
        )
        assert len(coefficients) == 3
        assert len(span_verification_conditions) == 0
        assert len(transformed_vars) == 2


def test_create_convex_hull_linear_inequalities_returns_correct_conditions_with_paper_example(convex_hull_learner: ConvexHullLearner,):
    pre_state_data = {
        "(x)": [1, 0, 0],
        "(y)": [0, 1, 0],
        "(z)": [0, 0, 1],
    }
    pre_state_df = DataFrame(pre_state_data)
    (coefficients, border_point, transformed_vars, span_verification_conditions,) = convex_hull_learner._create_convex_hull_linear_inequalities(
        pre_state_df, display_mode=False
    )
    print(span_verification_conditions)
    assert len(span_verification_conditions) == 1  # One dimension was reduced.
    assert len(coefficients) == 3
    assert len(transformed_vars) == 2


def test_create_convex_hull_linear_creates_convex_hull_with_no_duplicate_equations_due_to_numeric_precision_restriction(
    convex_hull_learner: ConvexHullLearner,
):
    pre_state_data = {
        "(x)": [0, 0, 0, 7, 7, 1.2300056, 1.2300057],
        "(y)": [0, 1, 0, 0, 7, 2.5600057, 2.5600058],
        "(z)": [0, 0, 1, 0, 7, 3.8900058, 3.8900059],
    }
    pre_state_df = DataFrame(pre_state_data)
    (coefficients, border_point, transformed_vars, span_verification_conditions,) = convex_hull_learner._create_convex_hull_linear_inequalities(
        pre_state_df, display_mode=False
    )
    assert len(span_verification_conditions) == 0
    assert len([str(c) for c in coefficients]) == len(set([str(c) for c in coefficients]))


def test_create_convex_hull_with_and_without_linear_dependency_removal_should_create_the_same_number_of_inequalities_when_there_are_constants_and_linear_dependent_features(
    convex_hull_learner: ConvexHullLearner,
):
    # we have data with 4 features, one feature is constant with the value of 1, and the other two are linearly dependent on one another and the last one is independent.
    # we create 10 samples of this data, and then we create the convex hull with and without removing the linear dependency.
    # The number of inequalities should be the same.
    NUM_SAMPLES = 10
    pre_state_data = {
        "(x )": [1] * NUM_SAMPLES,
        "(y )": [i for i in range(NUM_SAMPLES)],
        "(z )": [2 * i for i in range(NUM_SAMPLES)],
        "(w )": [random.randint(0, 100) for _ in range(NUM_SAMPLES)],
    }
    pre_state_df = DataFrame(pre_state_data)
    filtered_df, linear_deps = remove_complex_linear_dependencies(pre_state_df)
    assert len(filtered_df.columns.tolist()) < len(pre_state_df.columns.tolist())
    (coefficients, border_point, transformed_vars, span_verification_conditions,) = convex_hull_learner._create_convex_hull_linear_inequalities(
        pre_state_df, display_mode=False
    )
    (
        coefficients_filtered,
        border_point_filtered,
        transformed_vars_filtered,
        span_verification_conditions_filtered,
    ) = convex_hull_learner._create_convex_hull_linear_inequalities(filtered_df, display_mode=False)
    assert len(coefficients_filtered) == len(coefficients)


def test_construct_safe_linear_inequalities_with_relevant_fluents_ignores_all_variables_that_are_not_in_the_relevant_fluents(
    convex_hull_learner: ConvexHullLearner,
):
    for i in range(10):
        convex_hull_learner.add_new_point(
            {"(x )": random.randint(0, 100), "(y )": random.randint(0, 100), "(z )": random.randint(0, 100),}
        )
    inequality_precondition = convex_hull_learner.construct_safe_linear_inequalities(["(x )"])
    print(str(inequality_precondition))
    assert inequality_precondition.binary_operator == "and"
    assert len(inequality_precondition.operands) == 2  # One for upper bound and one for lower bound.
    for operand in inequality_precondition.operands:
        assert "(x )" in operand.to_pddl()
        assert "(y )" not in operand.to_pddl()
        assert "(z )" not in operand.to_pddl()


def test_epsilon_approximate_hull_when_given_epsilon_zero_and_no_parameters_returns_original_convex_hull(convex_hull_learner: ConvexHullLearner,):
    # create a three-dimensional np.array with 10 samples of random numbers.
    data = np.random.randint(0, 100, (10, 3))
    convex_hull_learner._epsilon = 0
    convex_hull_learner._qhull_options = ""
    non_approximated_hull = convex_hull_learner._epsilon_approximate_hull(data)
    expected_convex_hull = ConvexHull(data)
    assert np.equal(non_approximated_hull.points, expected_convex_hull.points).all()
    assert np.equal(non_approximated_hull.equations, expected_convex_hull.equations).all()


def test_construct_safe_linear_inequalities_with_example_from_the_new_paper_returns_correct_values(convex_hull_learner: ConvexHullLearner,):
    pre_state_data = {
        "(x)": [2, 1, 11],
        "(y)": [0, 0, 0],
        "(z)": [1, 1, 0],
    }
    pre_state_df = DataFrame(pre_state_data)
    (coefficients, border_point, transformed_vars, span_verification_conditions,) = convex_hull_learner._create_convex_hull_linear_inequalities(
        pre_state_df, display_mode=False
    )
    print(coefficients)



from collections import defaultdict
from typing import Dict, Set, List

from pddl_plus_parser.models import Action, Predicate

from sam_learning.core import LearnerAction


def calculate_number_true_positives(learned_predicates: Set[str], expected_predicates: Set[str]) -> int:
    """Calculates the number of predicates that appear both in the model domain and in the learned domain.

    :param learned_predicates: the predicates that belong in the learned domain.
    :param expected_predicates: the predicates that belong to the model domain.
    :return: the number of predicates that match in both domains.
    """
    return len(learned_predicates.intersection(expected_predicates))


def calculate_number_false_positives(learned_predicates: Set[str], expected_predicates: Set[str]) -> int:
    """Calculates the number of predicates that appear in the learned domain but are not in the model domain.

    :param learned_predicates: the predicates that belong in the learned domain.
    :param expected_predicates: the predicates that belong to the model domain.
    :return: the number of predicates that belong to the learned domain and not the model domain.
    """
    return len(learned_predicates.difference(expected_predicates))


def calculate_number_false_negatives(learned_predicates: Set[str], expected_predicates: Set[str]) -> int:
    """Calculates the number of predicates that are missing in the learned domain from the model domain.

    :param learned_predicates: the predicates that belong in the learned domain.
    :param expected_predicates: the predicates that belong to the model domain.
    :return: the number of predicates missing in the learned domain.
    """
    return len(expected_predicates.difference(learned_predicates))


def calculate_syntactic_recall(learned_predicates: Set[str], actual_predicates: Set[str]) -> float:
    """Calculates the recall value of the input predicates.

    :param learned_predicates: the predicates learned using the learning algorithm.
    :param actual_predicates: the predicates belonging to the model domain.
    :return: the recall value.
    """
    if len(actual_predicates) == 0:
        return 1

    if len(learned_predicates) == 0:
        return 0

    true_positives = calculate_number_true_positives(learned_predicates, actual_predicates)
    false_negatives = calculate_number_false_negatives(learned_predicates, actual_predicates)
    return true_positives / (true_positives + false_negatives)


def calculate_syntactic_precision(learned_predicates: Set[str], actual_predicates: Set[str]) -> float:
    """Calculates the precision value of the input predicates.

    :param learned_predicates: the predicates learned using the learning algorithm.
    :param actual_predicates: the predicates belonging to the model domain.
    :return: the precision value.
    """
    if len(learned_predicates) == 0:
        if len(actual_predicates) == 0:
            return 1

        return 0

    true_positives = calculate_number_true_positives(learned_predicates, actual_predicates)
    false_positives = calculate_number_false_positives(learned_predicates, actual_predicates)
    return true_positives / (true_positives + false_positives)


class PrecisionRecallCalculator:
    """Class that manages the calculation of the precision and recall of the learned model."""

    preconditions: Dict[str, Set[str]]
    ground_truth_preconditions: Dict[str, Set[str]]
    discrete_effects: Dict[str, Set[str]]
    ground_discrete_effects: Dict[str, Set[str]]
    _learned_actions: List[str]

    def __init__(self):
        self.preconditions = defaultdict(set)
        self.ground_truth_preconditions = defaultdict(set)
        self.discrete_effects = defaultdict(set)
        self.ground_discrete_effects = defaultdict(set)
        self._compared_tuples = [(self.preconditions, self.ground_truth_preconditions), (self.discrete_effects, self.ground_discrete_effects)]
        self._learned_actions = []

    def add_action_data(self, learned_action: LearnerAction, model_action: Action) -> None:
        """Adds the discrete action's data to the class data to calculate the precision and recall values.

        :param learned_action: the action that was learned using the action model learning algorithm.
        :param model_action: the expected action of the original domain.
        """
        self._learned_actions.append(learned_action.name)
        self.preconditions[learned_action.name] = {p.untyped_representation for _, p in learned_action.preconditions if isinstance(p, Predicate)}

        self.ground_truth_preconditions[model_action.name] = {
            p.untyped_representation for _, p in model_action.preconditions if isinstance(p, Predicate)
        }

        self.discrete_effects[learned_action.name] = {p.untyped_representation for p in learned_action.discrete_effects}
        self.ground_discrete_effects[model_action.name] = {p.untyped_representation for p in model_action.discrete_effects}

    def calculate_action_precision(self, action_name: str) -> float:
        """calculates the precision value of a certain action.

        :param action_name: the name of the action that is being tested.
        :return: the action's precision.
        """
        true_positives = sum(calculate_number_true_positives(tup[0][action_name], tup[1][action_name]) for tup in self._compared_tuples)
        false_positives = sum(calculate_number_false_positives(tup[0][action_name], tup[1][action_name]) for tup in self._compared_tuples)
        if true_positives == 0 and false_positives == 0:
            return 1

        return true_positives / (true_positives + false_positives)

    def calculate_action_recall(self, action_name: str) -> float:
        """calculates the recall value of a certain action.

        :param action_name: the name of the action that is being tested.
        :return: the action's recall.
        """
        true_positives = sum(calculate_number_true_positives(tup[0][action_name], tup[1][action_name]) for tup in self._compared_tuples)
        false_negatives = sum(calculate_number_false_negatives(tup[0][action_name], tup[1][action_name]) for tup in self._compared_tuples)

        if true_positives == 0 and false_negatives == 0:
            return 1

        return true_positives / (true_positives + false_negatives)

    def calculate_syntactic_precision(self) -> float:
        """calculates the precision value of a learned domain.

        :return: the model's precision.
        """
        true_positives = sum(
            [
                sum(calculate_number_true_positives(tup[0][action_name], tup[1][action_name]) for tup in self._compared_tuples)
                for action_name in self._learned_actions
            ]
        )
        false_positives = sum(
            [
                sum(calculate_number_false_positives(tup[0][action_name], tup[1][action_name]) for tup in self._compared_tuples)
                for action_name in self._learned_actions
            ]
        )
        if true_positives == 0 and false_positives == 0:
            return 1

        return true_positives / (true_positives + false_positives)

    def calculate_syntactic_recall(self) -> float:
        """calculates the recall value of a learned domain.

        :return: the model's recall.
        """
        true_positives = sum(
            [
                sum(calculate_number_true_positives(tup[0][action_name], tup[1][action_name]) for tup in self._compared_tuples)
                for action_name in self._learned_actions
            ]
        )
        false_negatives = sum(
            [
                sum(calculate_number_false_negatives(tup[0][action_name], tup[1][action_name]) for tup in self._compared_tuples)
                for action_name in self._learned_actions
            ]
        )
        if true_positives == 0 and false_negatives == 0:
            return 1

        return true_positives / (true_positives + false_negatives)

    def export_action_syntactic_statistics(self, action_name: str) -> Dict[str, float]:
        """Export the statistics of an action that was observed during the learning process.

        :param action_name: the name of the action that was learned.
        :return: the dictionary containing the precision and recall statistics of the action.
        """
        action_precision = self.calculate_action_precision(action_name)
        action_recall = self.calculate_action_recall(action_name)
        action_f1_score = (
            0 if action_precision == 0 and action_recall == 0 else 2 * (action_precision * action_recall) / (action_precision + action_recall)
        )

        return {
            "preconditions_precision": calculate_syntactic_precision(self.preconditions[action_name], self.ground_truth_preconditions[action_name]),
            "effects_precision": calculate_syntactic_precision(self.discrete_effects[action_name], self.ground_discrete_effects[action_name]),
            "preconditions_recall": calculate_syntactic_recall(self.preconditions[action_name], self.ground_truth_preconditions[action_name]),
            "effects_recall": calculate_syntactic_recall(self.discrete_effects[action_name], self.ground_discrete_effects[action_name]),
            "action_precision": action_precision,
            "action_recall": action_recall,
            "f1_score": action_f1_score,
        }



"""Create a train and test set split for the input directory for the action model learning algorithms to use."""
import logging
import random
import shutil
from pathlib import Path
from typing import Tuple, List, Optional, Any, Dict

from .k_fold_split import create_test_set_indices, save_fold_settings, FOLDS_LABEL, load_distributed_fold_settings


class DistributedKFoldSplit:
    """Split the working directory into a train and test set directories."""

    logger: logging.Logger
    working_directory_path: Path
    n_split: int
    train_set_dir_path: Path
    test_set_dir_path: Path
    domain_file_path: Path
    only_train_test: bool

    def __init__(
        self,
        working_directory_path: Path,
        domain_file_name: str,
        n_split: int = 0,
        only_train_test: bool = False,
        learning_algorithms: List[int] = None,
        internal_iterations: List[int] = None,
    ):
        self.logger = logging.getLogger(__name__)
        self.working_directory_path = working_directory_path
        self.n_split = n_split
        self.train_set_dir_path = working_directory_path / "train"
        self.test_set_dir_path = working_directory_path / "test"
        self.domain_file_path = working_directory_path / domain_file_name
        self.only_train_test = only_train_test
        self._learning_algorithms = learning_algorithms
        self._internal_iterations = internal_iterations

    def _copy_domain(self, directory_path: Path) -> None:
        """Copies the domain to the train set directory so that it'd be used in the learning process."""
        self.logger.debug("Copying the domain to the train set directory.")
        shutil.copy(self.domain_file_path, directory_path / self.domain_file_path.name)

    def remove_created_directories(self) -> None:
        """Deletes the train and test directories."""

        self.logger.debug("Deleting the train set directory!")
        shutil.rmtree(self.train_set_dir_path, ignore_errors=True)
        self.logger.debug("Deleting the test set directory!")
        shutil.rmtree(self.test_set_dir_path, ignore_errors=True)

    def create_directories_content(
        self,
        test_set_problems: List[Path],
        selected_training_trajectories: Optional[List[Path]],
        fold_index: int,
        learning_algorithm: int,
        internal_iteration: Optional[int] = None,
        additional_suffix: str = "",
    ) -> Tuple[Path, Path]:
        """Creates the content of the train and test set directories."""
        fold_train_dir_path = (
            self.train_set_dir_path / f"fold_{fold_index}_{learning_algorithm}{additional_suffix}"
            f"{f'_{internal_iteration}' if internal_iteration is not None else ''}"
        )
        fold_test_dir_path = (
            self.test_set_dir_path / f"fold_{fold_index}_{learning_algorithm}{additional_suffix}"
            f"{f'_{internal_iteration}{additional_suffix}' if internal_iteration is not None else ''}"
        )
        fold_train_dir_path.mkdir(exist_ok=True)
        fold_test_dir_path.mkdir(exist_ok=True)
        self._copy_domain(fold_train_dir_path)

        for problem in test_set_problems:
            shutil.copy(problem, fold_test_dir_path / problem.name)

        for trajectory in selected_training_trajectories:
            shutil.copy(trajectory, fold_train_dir_path / trajectory.name)
            problem_path = self.working_directory_path / f"{trajectory.stem}.pddl"
            shutil.copy(problem_path, fold_train_dir_path / problem_path.name)

        return fold_train_dir_path, fold_test_dir_path

    def create_fold_from_scratch(self, folds_data: Dict[str, Any], max_items: int, trajectory_suffix: str) -> None:
        trajectory_paths = list(self.working_directory_path.glob(trajectory_suffix))
        items_per_fold = max_items if (0 < max_items <= len(trajectory_paths)) else len(trajectory_paths)
        self.logger.debug(f"Creating {items_per_fold} items per fold.")
        trajectory_paths = random.sample(trajectory_paths, k=items_per_fold)
        for fold_index, (train_set_indices, test_set_indices) in enumerate(
            create_test_set_indices(items_per_fold, self.n_split, self.only_train_test)
        ):
            test_set_problems = [self.working_directory_path / f"{trajectory_paths[i].stem}.pddl" for i in test_set_indices]
            train_set_trajectories = [trajectory_paths[i] for i in range(len(trajectory_paths)) if i not in test_set_indices]
            training_data = {"internal_iterations": {}}
            for num_used_trajectories in self._internal_iterations:
                self.logger.debug(f"Creating fold {fold_index} with {num_used_trajectories} trajectories.")
                selected_trajectories = train_set_trajectories[:num_used_trajectories]
                for learning_algorithm in self._learning_algorithms:
                    self.create_directories_content(
                        test_set_problems=test_set_problems,
                        selected_training_trajectories=selected_trajectories,
                        fold_index=fold_index,
                        learning_algorithm=learning_algorithm,
                        internal_iteration=num_used_trajectories,
                    )

                training_data["internal_iterations"][int(num_used_trajectories)] = [str(p.absolute()) for p in selected_trajectories]

            max_dataset_size = self._internal_iterations[-1]
            for learning_algorithm in self._learning_algorithms:
                self.create_directories_content(
                    test_set_problems=test_set_problems,
                    selected_training_trajectories=train_set_trajectories[:max_dataset_size],
                    fold_index=fold_index,
                    learning_algorithm=learning_algorithm,
                    additional_suffix="_triplets",
                )

            folds_data[f"{FOLDS_LABEL}_{fold_index}"] = {
                "train": training_data,
                "test": [str(p.absolute()) for p in test_set_problems],
            }
            self.logger.debug("Finished creating fold!")

        save_fold_settings(self.working_directory_path, folds_data)

    def create_k_fold(self, trajectory_suffix: str = "*.trajectory", max_items: int = 0, load_configuration: bool = True):
        """Creates a generator that will be used for the next algorithm to know where the train and test set
            directories reside.

        :param trajectory_suffix: the suffix of the trajectory files to be used.
        :param max_items: the maximum number of items to be used in the train and test set together.
        :param load_configuration: whether to load folds settings from the configuration file.
        :return: a generator for the train and test set directories.
        """
        self.logger.info("Starting to create the folds for the cross validation process.")
        self.train_set_dir_path.mkdir(exist_ok=True)
        self.test_set_dir_path.mkdir(exist_ok=True)
        folds_data = load_distributed_fold_settings(self.working_directory_path)
        train_test_paths = []
        if load_configuration and len(folds_data) > 0:
            self.logger.debug("Loading the folds settings from the configuration file.")
            for index, (fold_name, fold_content) in enumerate(folds_data.items()):
                for iteration_num, selected_trajectories in fold_content["train"]["internal_iterations"].items():
                    selected_trajectories_paths = [Path(p) for p in selected_trajectories]
                    for learning_algorithm in self._learning_algorithms:
                        self.logger.debug("Creating fold directories content for each learning algorithm.")
                        train_test_paths.append(
                            self.create_directories_content(
                                test_set_problems=fold_content["test"],
                                selected_training_trajectories=selected_trajectories_paths,
                                fold_index=index,
                                learning_algorithm=learning_algorithm,
                                internal_iteration=int(iteration_num),
                            )
                        )

                self.logger.debug("Finished creating fold!")
                # creating the folder to be used for the triplets based experiments
                largest_iteration = max([int(i) for i in fold_content["train"]["internal_iterations"].keys()])
                relevant_paths = [Path(p) for p in fold_content["train"]["internal_iterations"][str(largest_iteration)]]
                for learning_algorithm in self._learning_algorithms:
                    self.create_directories_content(
                        test_set_problems=fold_content["test"],
                        selected_training_trajectories=relevant_paths,
                        fold_index=index,
                        learning_algorithm=learning_algorithm,
                        additional_suffix="_triplets",
                    )

            return

        self.create_fold_from_scratch(folds_data, max_items, trajectory_suffix)



"""Converts regular trajectories to the type of trajectories that PlanMiner algorithm accepts."""
import logging
import sys
from pathlib import Path
from typing import Set, Dict

from pddl_plus_parser.exporters import DomainExporter
from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser, TrajectoryParser
from pddl_plus_parser.models import Operator, GroundedPredicate, PDDLFunction, Domain

from sam_learning.core.environment_snapshot import EnvironmentSnapshot


class PlanMinerTrajectoriesCreator:
    """Class that transforms our trajectory file data into the type of trajectories that plan miner accepts.

    Note:
        This assumes that there are regular trajectories to be converted to the correct format.
    """

    domain_file_name: str
    working_directory_path: Path
    logger: logging.Logger

    def __init__(self, domain_file_name: str, working_directory_path: Path):
        self.domain_file_name = domain_file_name
        self.working_directory_path = working_directory_path
        self.logger = logging.getLogger(__name__)

    def convert_domain_to_plan_miner_form(self, domain: Domain) -> None:
        """Converts the domain to the format that PlanMiner later outputs.

        Note:
            This is a helper function for the experiments where the comparison between the learned domain and the
            original domain is needed.
        """
        export_path = self.working_directory_path / f"{self.domain_file_name.split('.')[0]}_plan_miner.pddl"
        # change the names of the parameters in the predicates to be in the form of ?param_<index>
        for predicate in domain.predicates.values():
            old_to_new_parameter_names = {param_name: f"?param_{index}" for index, param_name in enumerate(predicate.signature.keys())}
            predicate.change_signature(old_to_new_parameter_names)

        # change the names of the parameters in the numeric functions to be in the form of ?param_<index>
        for function in domain.functions.values():
            old_to_new_parameter_names = {param_name: f"?param_{index}" for index, param_name in enumerate(function.signature.keys())}
            function.change_signature(old_to_new_parameter_names)

        # change the names of the parameters in the actions to be in the form of ?param_<index>
        for action in domain.actions.values():
            old_to_new_parameter_names = {param_name: f"?param_{index}" for index, param_name in enumerate(action.signature.keys())}
            action.change_signature(old_to_new_parameter_names)

        # export the new domain to a file
        domain_exporter = DomainExporter()
        domain_exporter.export_domain(domain=domain, export_path=export_path)

    @staticmethod
    def create_complete_state(state_predicates: Set[GroundedPredicate], state_functions: Dict[str, PDDLFunction]) -> str:
        """Creates the complete state string from the state predicates and functions in the format that PlanMiner requires.

        :param state_predicates: the positive and negative state predicates.
        :param state_functions: the numeric functions in the state.
        :return: the complete state string in the PlanMiner format.
        """
        positive_state_predicates = [str(predicate).upper() for predicate in state_predicates if predicate.is_positive]
        negative_state_predicates = []
        for predicate in state_predicates:
            if not predicate.is_positive:
                negated_predicate = GroundedPredicate(
                    name=predicate.name, signature=predicate.signature, object_mapping=predicate.object_mapping, is_positive=True
                )
                negative_state_predicates.append(f"(not {str(negated_predicate).upper()})")

        functions = [function.state_typed_representation.upper() for function in state_functions.values()]
        return f"({' '.join(positive_state_predicates)} {' '.join(functions)} {' '.join(negative_state_predicates)})"

    def create_plan_miner_trajectories(self) -> None:
        """Creates the trajectories in the format file that PlanMiner requires.."""
        domain_file_path = self.working_directory_path / self.domain_file_name
        domain = DomainParser(domain_file_path).parse_domain()
        self.convert_domain_to_plan_miner_form(domain)
        triplet_snapshot = EnvironmentSnapshot(domain)
        plan_miner_trajectory_file_path = self.working_directory_path / "plan_miner_trajectories.pts"
        if plan_miner_trajectory_file_path.exists():
            return

        trajectories = []
        for trajectory_file_path in self.working_directory_path.glob("*.trajectory"):
            self.logger.info(f"Processing trajectory file: {trajectory_file_path}")
            problem_file_path = self.working_directory_path / f"{trajectory_file_path.stem}.pddl"
            problem = ProblemParser(problem_file_path, domain).parse_problem()
            observation = TrajectoryParser(domain, problem).parse_trajectory(trajectory_file_path)
            plan_sequence = ["##Tasks##"]
            state_sequence = ["##States##"]
            for index, action_triplet in enumerate(observation.components):
                action = action_triplet.grounded_action_call
                prev_state = action_triplet.previous_state
                next_state = action_triplet.next_state
                previous_state_predicates = triplet_snapshot._create_state_discrete_snapshot(prev_state, observation.grounded_objects)
                next_state_predicates = triplet_snapshot._create_state_discrete_snapshot(next_state, observation.grounded_objects)
                op = Operator(
                    action=domain.actions[action.name], domain=domain, grounded_action_call=action.parameters, problem_objects=problem.objects
                )
                plan_sequence.append(f"[{index}, {index + 1}]: {op.typed_action_call.upper()}")
                if index == 0:
                    state_str = self.create_complete_state(state_predicates=previous_state_predicates, state_functions=prev_state.state_fluents)
                    state_sequence.append(f"[{index}]: {state_str}")

                state_str = self.create_complete_state(state_predicates=next_state_predicates, state_functions=next_state.state_fluents)
                state_sequence.append(f"[{index + 1}]: {state_str}")

            action_trace = "\n".join(plan_sequence)
            state_trace = "\n\n".join(state_sequence)
            plan_miner_trajectory = f"New plan!!!\n\n" f"{action_trace}\n\n\n" f"{state_trace}"
            trajectories.append(plan_miner_trajectory)

        with open(plan_miner_trajectory_file_path, "wt") as plan_miner_trajectory_file:
            plan_miner_trajectory_file.write("\n\n".join(trajectories))


if __name__ == "__main__":
    trajectory_creator = PlanMinerTrajectoriesCreator(sys.argv[1], Path(sys.argv[2]))
    trajectory_creator.create_plan_miner_trajectories()



"""Module that stores amd learns an action's numeric state fluents."""
import logging
from collections import defaultdict
from typing import Dict, List, Tuple, Optional, Set

import numpy as np
from pddl_plus_parser.models import PDDLFunction, Precondition, NumericalExpressionTree

from sam_learning.core import NotSafeActionError, EquationSolutionType
from sam_learning.core.baseline_algorithms_version.naive_convex_hull_learner import NaiveConvexHullLearner
from sam_learning.core.baseline_algorithms_version.naive_linear_regression_learner import NaiveLinearRegressionLearner

np.seterr(divide='ignore', invalid='ignore')


class NaiveNumericFluentStateStorage:
    """Stores and learned the numeric state fluents of a single action."""

    logger: logging.Logger
    previous_state_storage: Dict[str, List[float]]  # lifted function str -> numeric values.
    negative_samples_storage: Dict[str, List[float]]  # lifted function str -> numeric values.
    next_state_storage: Dict[str, List[float]]  # lifted function str -> numeric values.

    def __init__(self, action_name: str, domain_functions: Dict[str, PDDLFunction]):
        self.logger = logging.getLogger(__name__)
        self.previous_state_storage = defaultdict(list)
        self.next_state_storage = defaultdict(list)
        self.action_name = action_name
        self.convex_hull_learner = NaiveConvexHullLearner(action_name, domain_functions)
        self.linear_regression_learner = NaiveLinearRegressionLearner(action_name, domain_functions)

    def add_to_previous_state_storage(self, state_fluents: Dict[str, PDDLFunction]) -> None:
        """Adds the matched lifted state fluents to the previous state storage.

        :param state_fluents: the lifted state fluents that were matched for the action.
        """
        for state_fluent_lifted_str, state_fluent_data in state_fluents.items():
            self.previous_state_storage[state_fluent_lifted_str].append(state_fluent_data.value)

    def add_to_next_state_storage(self, state_fluents: Dict[str, PDDLFunction]) -> None:
        """Adds the matched lifted state fluents to the next state storage.

        :param state_fluents: the lifted state fluents that were matched for the action.
        """
        for state_fluent_lifted_str, state_fluent_data in state_fluents.items():
            self.next_state_storage[state_fluent_lifted_str].append(state_fluent_data.value)
            if len(self.previous_state_storage.get(state_fluent_lifted_str, [])) != \
                    len(self.next_state_storage[state_fluent_lifted_str]):
                self.logger.debug("This is a case where effects create new fluents - should adjust the previous state.")
                self.previous_state_storage[state_fluent_lifted_str].append(0)

    def filter_out_inconsistent_state_variables(self) -> None:
        """Filters out fluents that appear only in part of the states since they are not safe.

        :return: only the safe state variables that appear in *all* states.
        """
        max_function_len = max([len(values) for values in self.previous_state_storage.values()])
        self.previous_state_storage = {lifted_function: state_values for lifted_function, state_values in
                                       self.previous_state_storage.items() if len(state_values) == max_function_len}
        self.next_state_storage = {lifted_function: state_values for lifted_function, state_values in
                                   self.next_state_storage.items() if len(state_values) == max_function_len}

    def construct_safe_linear_inequalities(
            self, relevant_fluents: Optional[List[str]] = None) -> Precondition:
        """Constructs the linear inequalities strings that will be used in the learned model later.

        :return: The precondition that contains the linear inequalities.
        """
        return self.convex_hull_learner.construct_safe_linear_inequalities(
            self.previous_state_storage, relevant_fluents)

    def _validate_safe_solving(self) -> None:
        """Validates that it is safe to apply linear regression to solve the input equations."""
        for lifted_function in self.next_state_storage:
            num_variables = len(self.previous_state_storage)
            num_equations = len(self.next_state_storage[lifted_function])
            # validate that it is possible to solve linear equations at all.
            if num_equations < num_variables or num_equations == num_variables == 1:
                failure_reason = f"Cannot solve linear equations when too little input equations given."
                self.logger.warning(failure_reason)
                raise NotSafeActionError(
                    self.action_name, failure_reason, EquationSolutionType.not_enough_data)

    def construct_assignment_equations(
            self, allow_unsafe_learning: bool = False, relevant_fluents: Optional[List[str]] = None) -> Tuple[
        Set[NumericalExpressionTree], Optional[Precondition], bool]:
        """Constructs the assignment statements for the action according to the changed value functions.

        :param allow_unsafe_learning: whether to allow learning from unsafe data.
        :param relevant_fluents: the fluents that are part of the action's preconditions and effects.
        :return: the constructed assignment statements.
        """
        self._validate_safe_solving()
        return self.linear_regression_learner.construct_assignment_equations(
            self.previous_state_storage, self.next_state_storage, allow_unsafe_learning=allow_unsafe_learning, relevant_fluents=relevant_fluents)



import sys
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def plot_solo_results(results_directory_path: Path, file_template: str = "*solving_combined_statistics.csv", using_triplets: bool = False):
    """Plot the results of the experiments."""
    for file_path in results_directory_path.glob(file_template):
        df = pd.read_csv(file_path)

        # Define color-blind friendly palette
        color_palette = sns.color_palette("colorblind", n_colors=len(df["learning_algorithm"].unique()))
        line_styles = ["-", "--", "-.", ":", "--", "-.", ":"]
        marker_options = ["o", "s", "D", "v", "^", ">", "<", "p", "P", "*", "X", "d"]
        line_styles_iterator = list(line_styles)
        color_cycle = list(color_palette)
        markers = list(marker_options)

        # Group the data by 'num_trajectories', 'learning_algorithm' and calculate the mean and std of 'percent_ok'
        df = df[~df["learning_algorithm"].isin(["incremental_nsam", "naive_nsam_no_dependency_removal"])]  # Remove max_percent_ok from the plot
        group_by_columns = ["num_trajectories", "learning_algorithm"] if not using_triplets else ["num_trajectory_triplets", "learning_algorithm"]
        grouped_data = df.groupby(group_by_columns, as_index=False).agg(avg_percent_ok=("percent_ok", "mean"), std_percent_ok=("percent_ok", "std"),)
        labels = {
            "numeric_sam": "NSAM*",
            "naive_nsam": "NSAM",
            "plan_miner": "Plan Miner",
        }

        # Plotting
        sns.set(style="whitegrid")
        stand_alone_fig = plt.figure(figsize=(12, 8))

        # Plot a line for each learning algorithm
        for index, algo in enumerate(df["learning_algorithm"].unique()):
            algo_data = grouped_data[grouped_data["learning_algorithm"] == algo]
            plot_x_axis = algo_data["num_trajectories"] if not using_triplets else algo_data["num_trajectory_triplets"]

            plt.plot(
                plot_x_axis,
                algo_data["avg_percent_ok"],
                linestyle=line_styles_iterator[index],
                label=labels[algo],
                marker=markers[index],
                color=color_cycle[index],
                linewidth=8,
            )

            plt.fill_between(
                plot_x_axis,
                np.clip(algo_data["avg_percent_ok"] - algo_data["std_percent_ok"], 0, 100),
                np.clip(algo_data["avg_percent_ok"] + algo_data["std_percent_ok"], 0, 100),
                alpha=0.2,
            )

        plt.xlabel(f"# {'Trajectories' if not using_triplets else 'Triplets'}", fontsize=44)
        plt.ylabel("AVG % of solved", fontsize=44)
        plt.ylim(0, 100)
        plt.tick_params(axis="both", which="major", labelsize=46)
        plt.grid(True)
        plt.legend(fontsize=44)

        output_file_path = file_path.parent / f"{file_path.stem}_plot_solo.pdf"
        plt.tight_layout()  # Ensures no overlap between plot elements
        plt.savefig(output_file_path, bbox_inches="tight", dpi=300)


if __name__ == "__main__":
    results_path = Path(sys.argv[1])
    plot_solo_results(results_path)
